# coding=utf-8
"""
Utilities functions and tools
These functions depend on the globals.py module and can generate circular dependencies when imported.
"""
import logging
import os
import shutil
import sys
import time
import tkinter as tk
from datetime import datetime
from io import BytesIO
from tkinter import messagebox

import requests
from PIL import Image, ImageTk
from termcolor import colored

from UEVaultManager.lfs.utils import path_join
from UEVaultManager.tkgui.modules import globals as gui_g
from UEVaultManager.tkgui.modules.cls.ProgressWindowClass import ProgressWindow


def log_format_message(name: str, levelname: str, message: str) -> str:
    """
    Format the log message.
    :param name: the name of the logger.
    :param levelname: the level of log.
    :param message: the message to format.
    :return: the formatted message.
    """
    return f'[{name}] {levelname}: {message}'


def box_message(msg: str, level='info'):
    """
    Display a message box with the given message.
    :param msg: the message to display.
    :param level: the level of the message (info, warning, error).
    """
    level_lower = level.lower()
    if level_lower == 'warning':
        log_warning(msg)
        messagebox.showwarning(title=gui_g.s.app_title, message=msg)
    elif level_lower == 'error':
        messagebox.showerror(title=gui_g.s.app_title, message=msg)
        log_error(msg)
        # done in log_error
        # exit(1)
    else:
        log_info(msg)
        messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def box_yesno(msg: str) -> bool:
    """
    Display a YES/NO message box with the given message.
    :param msg: the message to display .
    :return:  True if the user clicked on Yes, False otherwise.
    """
    return messagebox.askyesno(title=gui_g.s.app_title, message=msg)


def box_okcancel(msg: str) -> bool:
    """
    Display an OK/CANCEL message box with the given message.
    :param msg: the message to display.
    :return:  True if the user clicked on Yes, False otherwise.
    """
    return messagebox.askokcancel(title=gui_g.s.app_title, message=msg)


def todo_message() -> None:
    """
    Display a message box with a message saying that the feature is not implemented yet.
    """
    msg = 'Not implemented yet'
    print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(print_msg)
    messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def from_cli_only_message(content='This feature is only accessible') -> None:
    """
    Display a message box with a message saying that the feature is only accessible when running the app using the UEVM cli command options.
    """
    msg = f'{content} when running these app using the UEVM cli command options. Once the UEVaultManager package installed, Type UEVaultManager -h for more help'
    print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(print_msg)
    messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def log_info(msg: str) -> None:
    """
    Log an info message.
    :param msg: the message to log.

    Note: It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'blue'))
        print(print_msg)


def log_debug(msg: str) -> None:
    """
    Log a debug message.
    :param msg: the message to log.

    Note:
        It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
        This message will only be logged if the debug mode is enabled.
    """
    if not gui_g.s.debug_mode:
        return
    if gui_g.UEVM_log_ref is not None:
        # ensure that the debug messages will be logged even if the log level not set to DEBUG in the cli
        log_level = gui_g.UEVM_log_ref.level
        if log_level == logging.DEBUG:
            gui_g.UEVM_log_ref.debug(msg)
        else:
            gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Debug', colored(msg, 'light_grey'))
        print(print_msg)


def log_warning(msg: str) -> None:
    """
    Log a warning message.
    :param msg: the message to log.

    Note: It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Warning', colored(msg, 'orange'))
        print(print_msg)


def log_error(msg: str) -> None:
    """
    Log an error message.
    :param msg: the message to log. Note that the app will exit after logging the message.

    Note: It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.error(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Error', colored(msg, 'red', 'bold'))
        print(print_msg)
    if gui_g.UEVM_gui_ref is not None:
        gui_g.UEVM_gui_ref.quit()
    sys.exit(1)


def resize_and_show_image(image: Image, canvas: tk.Canvas, scale: float = 1.0) -> None:
    """
    Resize the given image and display it in the given canvas.
    :param image: the image to display.
    :param canvas: the canvas to display the image in.
    :param scale: the scale to apply to the image.
    """
    # Resize the image while keeping the aspect ratio
    target_height = int(gui_g.s.preview_max_height * scale)
    aspect_ratio = float(image.width * scale) / float(image.height * scale)
    target_width = int(target_height * aspect_ratio)
    resized_image = image.resize((target_width, target_height), Image.BILINEAR)
    tk_image = ImageTk.PhotoImage(resized_image)
    # Calculate center coordinates
    x = max(0, (canvas.winfo_width() - tk_image.width()) // 2)
    y = max(0, (canvas.winfo_height() - tk_image.height()) // 2)
    canvas.create_image(x, y, anchor=tk.NW, image=tk_image)
    canvas.image = tk_image


def show_asset_image(image_url: str, canvas_image=None, scale: float = 1.0, timeout=(4, 4)) -> None:
    """
    Show the image of the given asset in the given canvas.
    :param image_url: the url of the image to display.
    :param canvas_image: the canvas to display the image in.
    :param scale: the scale to apply to the image.
    :param timeout: the timeout in seconds to wait for the image to be downloaded.
    """
    if canvas_image is None or image_url is None or not image_url or str(image_url) in ('', 'None', 'nan'):
        return
    try:
        # print(image_url)
        # noinspection DuplicatedCode
        if not os.path.isdir(gui_g.s.cache_folder):
            os.mkdir(gui_g.s.cache_folder)
        image_filename = path_join(gui_g.s.cache_folder, os.path.basename(image_url))
        # Check if the image is already cached
        if os.path.isfile(image_filename) and (time.time() - os.path.getmtime(image_filename)) < gui_g.s.image_cache_max_time:
            # Load the image from the cache folder
            image = Image.open(image_filename)
        else:
            response = requests.get(image_url, timeout=timeout)
            image = Image.open(BytesIO(response.content))

            with open(image_filename, "wb") as f:
                f.write(response.content)
        resize_and_show_image(image, canvas_image, scale)
    except Exception as error:
        log_warning(f"Error showing image: {error!r}")


def show_default_image(canvas_image=None) -> None:
    """
    Show the default image in the given canvas.
    :param canvas_image: the canvas to display the image in.
    """
    if canvas_image is None:
        return
    try:
        # Load the default image
        if os.path.isfile(gui_g.s.default_image_filename):
            def_image = Image.open(gui_g.s.default_image_filename)
            # noinspection PyTypeChecker
            resize_and_show_image(def_image, canvas_image)
    except Exception as error:
        log_warning(f"Error showing default image {gui_g.s.default_image_filename} cwd:{os.getcwd()}: {error!r}")


def json_print_key_val(json_obj, indent=4, print_result=True, output_on_gui=False) -> None:
    """
    Pretty prints a JSON object in a simple 'key: value' format.
    :param json_obj:  The JSON object to print.
    :param indent: The number of spaces to indent each level.
    :param print_result: Determines whether to print the result.
    :param output_on_gui: Determines whether to print the result on the GUI.
    :return: The pretty printed JSON object.
    """

    def _process(obj, level=0):
        lines = []
        indent_str = ' ' * indent * level

        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, (dict, list)):
                    lines.append(f'{indent_str}{key}:')
                    lines.extend(_process(value, level + 1))
                else:
                    lines.append(f'{indent_str}{key}: {value}')
        elif isinstance(obj, list):
            for item in obj:
                lines.extend(_process(item, level))
        else:
            lines.append(f'{indent_str}{obj}')

        return lines

    result = '\n'.join(_process(json_obj))

    if print_result:
        if output_on_gui and gui_g.display_content_window_ref is not None:
            gui_g.display_content_window_ref.display(result)
        else:
            print(result)


def custom_print(text='', keep_mode=True) -> None:
    """
    Print the given text on the GUI if it's available, otherwise print it on the console².
    :param text: the text to print.
    :param keep_mode: whether to keep the existing content when adding a new one.
    """
    if gui_g.display_content_window_ref is not None:
        gui_g.display_content_window_ref.display(content=text + '\n', keep_mode=keep_mode)
    else:
        print(text)


def get_tk_root(container) -> tk.Tk:
    """
    Get the root window.
    :param container:  the container window or object
    :return: the root window
    """
    root = None
    # get the root window to avoid creating multiple progress windows
    try:
        # a tk window chid class
        root = container.winfo_toplevel()
    except (AttributeError, tk.TclError):
        # an editableTable class
        try:
            root = container.get_container.winfo_toplevel()
        except (AttributeError, tk.TclError):
            pass
    return root


def show_progress(
    parent,
    text='Working...Please wait',
    width=500,
    height=120,
    max_value_l=0,
    show_progress_l=False,
    show_btn_stop_l=False,
    quit_on_close: bool = False,
    function: callable = None,
    function_parameters: dict = None
) -> ProgressWindow:
    """
    Show the progress window. If the progress window does not exist, it will be created.
    :param parent: The parent window.
    :param text: The text to display in the progress window.
    :param width: The width of the progress window.
    :param height: The height of the progress window.
    :param max_value_l: The maximum value of the progress bar.
    :param show_progress_l: Whether to show the progress bar.
    :param show_btn_stop_l: Whether to show the stop button.
    :param function: the function to execute.
    :param function_parameters: the parameters of the function.
    :param quit_on_close: whether to quit the application when the window is closed.
    :return: The progress window.
    It will create a new progress window if one does not exist and update parent._progress_window
    """
    root = get_tk_root(parent)
    if root and root.progress_window is None:
        pw = ProgressWindow(
            title=gui_g.s.app_title,
            parent=parent,
            icon=gui_g.s.app_icon_filename,
            width=width,
            height=height,
            show_btn_stop=show_btn_stop_l,
            show_progress=show_progress_l,
            max_value=max_value_l,
            quit_on_close=quit_on_close,
            function=function,
            function_parameters=function_parameters
        )
        root.progress_window = pw
    else:
        pw = root.progress_window
    pw.set_text(text)
    pw.set_activation(False)
    pw.update()
    return pw


def close_progress(parent) -> None:
    """
    Close the progress window.
    :param parent: The parent window.
    It accesses to the parent.progress_window property
    """
    root = get_tk_root(parent)
    if root and root.progress_window is not None:
        root.progress_window.close_window()
        root.progress_window = None


def create_file_backup(file_src: str, logger: logging.Logger = None, path: str = '') -> str:
    """
    Create a backup of a file.
    :param file_src: path to the file to back up.
    :param logger: the logger to use to display info. Could be None.
    :param path: the path to the config folder.
    :return: the full path to the backup file.
    """
    # for files defined relatively to the config folder
    file_src = file_src.replace('~/.config', path)
    file_backup = ''
    if not file_src:
        return ''
    try:
        file_name_no_ext, file_ext = os.path.splitext(file_src)
        file_backup = f'{file_name_no_ext}_{datetime.now().strftime("%y-%m-%d_%H-%M-%S")}{file_ext}.BAK'
        shutil.copy(file_src, file_backup)
        if logger is not None:
            logger.info(f'File {file_src} has been copied to {file_backup}')
    except FileNotFoundError:
        if logger is not None:
            logger.info(f'File {file_src} has not been found')
    return file_backup


def update_loggers_level(logger: logging.Logger = None, debug_value=None) -> None:
    """
    Change the logger level of debug depending on the debug mode.
    :param logger: the logger
    :param debug_value: the value to set. If None, it will use the value of gui_g.s.debug_mode

    Note:
        Will also update all the loggers level of the UEVM classes.
        Call this function when the debug mode is changed.
    """
    if logger is not None:
        if logger.name not in gui_g.UEVM_logger_names:
            gui_g.UEVM_logger_names.append(logger.name)
    debug_value = gui_g.s.debug_mode if debug_value is None else debug_value
    for logger_name in gui_g.UEVM_logger_names:
        logger = logging.getLogger(logger_name)
        logger.setLevel(level=logging.DEBUG if debug_value else logging.INFO)


def make_modal(window: tk.Toplevel = None, wait_for_close=True) -> None:
    """
    Make the given window modal.
    :param window: the window to make modal
    :param wait_for_close: whether to wait for the window to be closed before continuing
    """
    window.grab_set()
    window.focus_set()
    if wait_for_close:
        # Note: this will block the main window
        window.wait_window()



# coding=utf-8
"""
Utilities functions and tools
These functions DO NOT depend on the globals.py module and be freely imported.
"""
import ctypes as ct
import datetime
import os
import subprocess
import sys
import uuid

import ttkbootstrap as ttk
from screeninfo import get_monitors
from ttkbootstrap.publisher import Publisher

from UEVaultManager.lfs.utils import path_join


def log(message: str) -> None:
    """
    Log a message in the console.
    :param message: the message to log.
    """
    # keep this function as simple as possible to avoid circular import
    print(message)


def path_from_relative_to_absolute(path: str) -> str:
    """
    Build the path of the file to reference relative to the currently running script.
    :param path: the relative path to the file. If the path is already absolute, it is returned as is.
    :return: the absolute path of the file.
    """
    if os.path.isabs(path):
        return path

    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        # noinspection PyProtectedMember,PyUnresolvedReferences
        # base_path = sys._MEIPASS
        current_script_directory = sys._MEIPASS
        # when launched from pyinstaller, the path has been copied at the root of the folder where the script is run. So we nne to remove all the relative path indicator
        path = os.path.basename(path)
    except AttributeError:
        # base_path = os.path.abspath(".")
        current_script_path = os.path.abspath(__file__)
        current_script_directory = os.path.dirname(current_script_path)

    absolute_path = path_join(current_script_directory, path)
    return absolute_path


def center_window_on_screen(screen_index: int, width: int, height: int) -> str:
    """
    Calculate the geometry of the window to display in the center of the given screen.
    :param screen_index: the index of the screen to use.
    :param width: the width of the window.
    :param height: the height of the window.
    :return: the geometry string to use to display the window in the center of the screen.
    """
    x, y = get_center_screen_positions(screen_index, width, height)
    geometry: str = f'{width}x{height}+{x}+{y}'
    return geometry


def get_center_screen_positions(screen_index: int, width: int, height: int) -> (int, int):
    """
    Return the x and y positions of the window to display in the center of the given screen.
    :param screen_index: the index of the screen to use.
    :param width: the width of the window.
    :param height: the height of the window.
    :return: the geometry string to use to display the window in the center of the screen.
    """
    monitors = get_monitors()
    if screen_index > len(monitors):
        log(f'The screen #{screen_index} is not available. Using 0 as screen index.')  # no use of log functions here to prevent circular import
        screen_index = 0
    # Position the window in the center of the screen
    target_screen = monitors[screen_index]
    screen_width = target_screen.width
    screen_height = target_screen.height
    x = target_screen.x + (screen_width - width) // 2
    y = target_screen.y + (screen_height - height) // 2
    return x, y


def set_custom_style(theme_name='lumen', font=('Arial', 10, 'normal')):
    """
    Set the custom style for the application.
    :return: the style object.
    """
    try:
        style = ttk.Style(theme_name)
    except (Exception, ):
        # free the subscriber widgets to the Style. As it, it could be used by other window without issue
        Publisher.clear_subscribers()
        style = ttk.Style(theme_name)

    # option possible for ttk widgets:
    # TButton, TCheckbutton, TCombobox, TEntry, TFrame, TLabel, TLabelFrame, TMenubutton, TNotebook, TProgressbar, TRadiobutton,
    # TScale, TScrollbar, TSeparator, TSizegrip, Treeview, TPanedwindow,
    # Horizontal.TProgressbar or Vertical.TProgressbar (depending on the orient option),
    # Horizontal.TScale or Vertical.TScale (depending on the orient option),
    # Horizontal.TScrollbar or Vertical.TScrollbar (depending on the orient option)
    style.configure('TLabel', font=font, spacing=1, padding=2)
    style.configure('TButton', font=font, spacing=1, padding=2)
    style.configure('TEntry', font=font, spacing=1, padding=2)
    style.configure('TFrame', font=font, spacing=1, padding=1)
    style.configure('TCombobox', font=font, spacing=1, padding=1)
    style.configure('TLabelFrame', font=font, spacing=1, padding=1)
    return style


def set_toolbar_style(tk_window) -> None:
    """
    Remove the minimize and maximize buttons from a tkinter window.
    This version is compatible with Windows AND Non-windows OS
    # see https://stackoverflow.com/questions/2969870/removing-minimize-maximize-buttons-in-tkinter.
    :param tk_window: the tkinter window.
    """
    try:
        set_window_pos = ct.windll.user32.SetWindowPos
        set_window_long = ct.windll.user32.SetWindowLongPtrW
        get_window_long = ct.windll.user32.GetWindowLongPtrW
        get_parent = ct.windll.user32.GetParent
    except AttributeError:
        # Non-windows OS
        log('Non-windows OS detected. No need to remove the minimize and maximize buttons from the window.')
        return
    # Identifiers
    gwl_style = -16
    ws_minimizebox = 131072
    ws_maximizebox = 65536
    swp_nozorder = 4
    swp_nomove = 2
    swp_nosize = 1
    swp_framechanged = 32
    hwnd = get_parent(tk_window.winfo_id())
    old_style = get_window_long(hwnd, gwl_style)  # Get the style
    new_style = old_style & ~ws_maximizebox & ~ws_minimizebox  # New style, without max/min buttons
    set_window_long(hwnd, gwl_style, new_style)  # Apply the new style
    set_window_pos(hwnd, 0, 0, 0, 0, 0, swp_nomove | swp_nosize | swp_nozorder | swp_framechanged)  # Updates


def set_icon_and_minmax(tk_window, icon=None) -> None:
    """
    Set the icon and remove the min/max buttons of the window if no icon is provided.
    :param tk_window:.
    :param icon:
    """
    if icon is None:
        # remove the min/max buttons of the window
        # this code works on Window only
        # tk_window.attributes('-toolwindow', True)
        tk_window.after(300, lambda: set_toolbar_style(tk_window))
    else:
        # windows only (remove the minimize/maximize buttons and the icon)
        icon = path_from_relative_to_absolute(icon)
        if icon != '' and os.path.isfile(icon):
            try:
                tk_window.iconbitmap(icon)
            except Exception as error:
                # in linux, the ico can exist but not be readable
                log(f'Error while setting the icon: {error!r}')


def create_empty_file(file_path: str) -> (bool, str):
    """
    Create an empty file.
    :param file_path: the path of the file to create.
    :return: (True if path was valid, the corrected path of the file).
    """
    path, file = os.path.split(file_path)
    is_valid, path = check_and_get_folder(path)
    file_path = path_join(path, file)
    open(file_path, 'w').close()
    return is_valid, file_path


def check_and_get_folder(folder_path: str) -> (bool, str):
    """
    Check if the folder exists. If not, create it or use the default one.
    :param folder_path: the path of the folder to check.
    :return: (True if path was valid, the corrected path of the folder).
    """
    path = folder_path
    is_valid = True
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except (OSError, PermissionError) as e:
            is_valid = False
            log(f'Error while creating the directory {path}: {e}')
            if home_dir := os.environ.get('XDG_CONFIG_HOME'):
                path = path_join(home_dir, 'UEVaultManager')
            else:
                path = os.path.expanduser('~/.config/UEVaultManager')
            if not os.path.exists(path):
                os.makedirs(path)
                path = os.path.normpath(path)
            log(f'The following folder {path} will be used as default')

    path = os.path.normpath(path)
    return is_valid, path


def convert_to_bool(value) -> bool:
    """
    Convert a value to a boolean. Useful for None values.
    :param value: the value to convert. If the value is not a boolean, it will be converted to a string and then to a boolean.
    :return:
    """
    try:
        if str(value).lower() in ('1', '1.0', 'true', 'yes', 'y', 't'):
            return True
        else:
            return False
    except (TypeError, ValueError):
        return False


def convert_to_int(value, prefix: str = '') -> int:
    """
    Convert a value to an integer.
    :param value: the value to convert.
    :param prefix: a prefix to remove from the value before converting it to an integer.
    :return: the integer value or 0 if the value is None or not an integer.
    """
    # remove prefix if any
    if prefix and isinstance(value, str) and value.startswith(prefix):
        value = value[len(prefix):]
    try:
        value = int(value)
        return value
    except (TypeError, ValueError):
        return 0


def convert_to_float(value) -> float:
    """
    Convert a value to a float. Useful for None values.
    :param value: the value to convert.
    :return: the float value or 0.0 if the value is None or not a float.
    """
    try:
        value = float(value)
        return value
    except (TypeError, ValueError):
        return 0.0


def convert_to_datetime(value: str, formats_to_use='%Y-%m-%d %H:%M:%S', default=None) -> datetime.datetime:
    """
    Convert a value to a datetime object.
    :param value: the value to convert. If the value is not a datetime, it will be converted to a string and then to a datetime.
    :param formats_to_use: list of format to use to trye to convert the value. They will be tried in order.
    :param default: the default value to return if the conversion fails. If None, the default value is 1970-01-01 00:00:00.
    :return: the datetime value.
    """
    if default is None:
        default = datetime.datetime(1970, 1, 1)
    if not value:
        return default
    if not isinstance(formats_to_use, list):
        formats_to_use = [formats_to_use]
    for date_format in formats_to_use:
        try:
            value_date = datetime.datetime.strptime(value, date_format)
            if value_date != datetime.datetime(1970, 1, 1):
                return value_date
        except (TypeError, ValueError, AttributeError):
            continue

    return default


def convert_to_str_datetime(value, date_format='%Y-%m-%d %H:%M:%S', default=None) -> str:
    """
    Convert a value to a datetime string.
    :param value: the value to convert. If the value is not a datetime.
    :param date_format: the format of value.
    :param default: the default value to return if the conversion fails. If None, the default value is 1970-01-01 00:00:00.
    :return: the string value of the datetime.
    """
    # we convert only datetime object
    if isinstance(value, str):
        return value

    if default is None:
        default = datetime.datetime(1970, 1, 1).strftime(date_format)
    if not value:
        return default
    try:
        return value.strftime(date_format)
    except (TypeError, ValueError, AttributeError):
        return default


def is_an_int(value, prefix: str = '') -> bool:
    """
    Check if a value is an integer.
    :param value: the value to check.
    :param prefix: a prefix to remove from the value before checking if it is an integer.
    :return:  True if the value is an integer, False otherwise.
    """
    # remove prefix if any
    if prefix and isinstance(value, str) and value.startswith(prefix):
        value = value[len(prefix):]
    try:
        float_n = float(value)
        int_n = int(float_n)
    except ValueError:
        return False
    else:
        return float_n == int_n


def create_uid() -> str:
    """
    Create a unique id.
    :return: a unique id.
    """
    return str(uuid.uuid4())[:8]


def shorten_text(url: str, limit: int = 30) -> str:
    """
    Shorten an url. Get its last part
    :param url:  the url to shorten
    :param limit: the limit of characters to keep
    :return: the shortened url
    """
    if len(url) < limit:
        return url
    else:
        return '...' + url[-limit:]


def extract_variables_from_url(url: str) -> dict:
    """
    Extract variables from an url.
    :param url: the url to extract variables from
    :return: a dict containing the variables
    """
    result = {}
    url_parts = url.split('?')
    if len(url_parts) == 2:
        url_params = url_parts[1].split('&')
        extracted_data = {}
        for param in url_params:
            key, value = param.split('=')
            extracted_data[key] = value
        result = extracted_data
    return result


def open_folder_in_file_explorer(folder_path) -> bool:
    """
    Open a folder in the file explorer.
    :param folder_path: the path of the folder to open.
    """
    if os.path.exists(folder_path):
        try:
            # For Windows
            if os.name == 'nt':
                # if we use check = True, it will raise an error even if the folder is opened, because this command always return 1 on windows
                process = subprocess.run(['explorer', folder_path], check=False)
                result = process.returncode == 0 or process.returncode == 1
            # For macOS
            elif os.name == 'posix':
                process = subprocess.run(['open', folder_path], check=False)
                result = process.returncode == 0
            # For Linux
            else:
                process = subprocess.run(['xdg-open', folder_path], check=False)
                result = process.returncode == 0
            return result
        except subprocess.CalledProcessError as error:
            log(f'Failed to open {folder_path} in file explorer.Error code: {error!r}')
            return False
    else:
        return False



# coding: utf-8
"""
Implementation for:
- EPCAPI : Epic Games Client API
- GrabResult : Enum for the result of grabbing a page.
- create_empty_assets_extra : Creates an empty asset extra dict.
"""
import json
import logging
import re
from enum import Enum

import requests
import requests.adapters
from bs4 import BeautifulSoup
from requests.auth import HTTPBasicAuth

from UEVaultManager.models.exceptions import InvalidCredentialsError
from UEVaultManager.utils.cli import create_list_from_string


class GrabResult(Enum):
    """
    Enum for the result of grabbing a page.
    """
    NO_ERROR = 0
    # next codes could occur only with beautifulsoup data grabbing (UEVM Version 1.X.X.X)
    INCONSISTANT_DATA = 1
    PAGE_NOT_FOUND = 2
    CONTENT_NOT_FOUND = 3
    TIMEOUT = 4
    # next codes could occur only with API scraping only (UEVM version 2.X.X.X)
    PARTIAL = 5  # when asset has been added when owned asset data only (less complete that "standard" asset data)
    NO_APPID = 6  # no appid found in the data (will produce a file name like '_no_appId_asset_1e10acc0cca34d5c8ff7f0ab57e7f89f
    NO_RESPONSE = 7  # the url does not return HTTP 200


def is_asset_obsolete(supported_versions='', engine_version_for_obsolete_assets=None) -> bool:
    """
    :param supported_versions: The supported versions the check the obsolete status against.
    :param engine_version_for_obsolete_assets: The engine version to use to check if an asset is obsolete.
    :return: True if the asset is obsolete, False otherwise.
    """
    if not engine_version_for_obsolete_assets or not supported_versions:
        obsolete = False
    else:
        supported_versions_list = supported_versions.lower().replace('ue_', '')
        supported_versions_list = create_list_from_string(supported_versions_list)
        obsolete = True
        for _, version in enumerate(supported_versions_list):
            if version == '':
                continue
            else:
                if float(engine_version_for_obsolete_assets) <= float(version):
                    obsolete = False
                    break
    return obsolete


def create_empty_assets_extra(asset_name: str) -> dict:
    """
    Creates an empty asset extra dict.
    :param asset_name:  The name of the asset.
    :return: The empty asset extra dict.
     """
    return {
        'asset_name': asset_name,
        'asset_slug': '',
        'price': 0,
        'discount_price': 0,
        'review': 0,
        'owned': False,
        'discount_percentage': 0,
        'discounted': False,
        'asset_url': '',
        'page_title': '',
        'supported_versions': '',
        'grab_result': GrabResult.NO_ERROR.name,
    }


class EPCAPI:
    """
    Epic Games Client API.
    :param lc: The language code.
    :param cc: The country code.
    :param timeout: The timeout for requests.
    """
    ignored_logger = None

    _user_agent = 'UELauncher/11.0.1-14907503+++Portal+Release-Live Windows/10.0.19041.1.256.64bit'
    _store_user_agent = 'EpicGamesLauncher/14.0.8-22004686+++Portal+Release-Live'
    # required for the oauth request
    _user_basic = '34a02cf8f4414e29b15921876da36f9a'
    _pw_basic = 'daafbccc737745039dffe53d94fc76cf'
    _label = 'Live-EternalKnight'

    _oauth_host = 'account-public-service-prod03.ol.epicgames.com'
    _launcher_host = 'launcher-public-service-prod06.ol.epicgames.com'
    _entitlements_host = 'entitlement-public-service-prod08.ol.epicgames.com'
    _catalog_host = 'catalog-public-service-prod06.ol.epicgames.com'
    _ecommerce_host = 'ecommerceintegration-public-service-ecomprod02.ol.epicgames.com'
    _datastorage_host = 'datastorage-public-service-liveegs.live.use1a.on.epicgames.com'
    _library_host = 'library-service.live.use1a.on.epicgames.com'
    # Using the actual store host with a user-agent newer than 14.0.8 leads to a CF verification page,
    # but the dedicated graphql host works fine.
    # _store_gql_host = 'launcher.store.epicgames.com'
    _store_gql_host = 'graphql.epicgames.com'
    _artifact_service_host = 'artifact-public-service-prod.beee.live.use1a.on.epicgames.com'
    _login_url = 'www.unrealengine.com/id/login/epic'

    _url_marketplace = 'www.unrealengine.com/marketplace'
    _search_url = _url_marketplace + '/en-US'
    # _url_asset_list = 'https://www.unrealengine.com/marketplace/api/assets'
    _url_asset_list = _url_marketplace + '/api/assets'
    # _url_owned_assets = 'https://www.unrealengine.com/marketplace/api/assets/vault'
    _url_owned_assets = _url_asset_list + '/vault'
    # _url_asset = 'https://www.unrealengine.com/marketplace/api/assets/asset'
    _url_asset = _url_asset_list + '/asset'

    # page d'un asset avec son urlSlug
    # _url_marketplace/en-US/product/{'urlSlug}
    # https://www.unrealengine.com/marketplace/en-US/product/volcrate
    #
    # detail json d'un asset avec son id (et non pas son asset_id ou son catalog_id)
    # UE_ASSET/{el['id']}")
    # https://www.unrealengine.com/marketplace/api/assets/asset/d27cf128fdc24e328cf950b019563bc5
    #
    # liste json des reviews d'un asset avec son id
    # https://www.unrealengine.com/marketplace/api/review/d27cf128fdc24e328cf950b019563bc5/reviews/list?start=0&count=10&sortBy=CREATEDAT&sortDir=DESC
    #
    # liste json des questions d'un asset avec son id
    # https://www.unrealengine.com/marketplace/api/review/d27cf128fdc24e328cf950b019563bc5/questions/list?start=0&count=10&sortBy=CREATEDAT&sortDir=DESC
    #
    # liste json des tags courants
    # https://www.unrealengine.com/marketplace/api/tags

    def __init__(self, lc='en', cc='US', timeout=(7, 7)):
        self.log = logging.getLogger('EPCAPI')
        self.notfound_logger = None  # will be setup when created in core.py
        self.session = requests.session()
        self.session.headers['User-Agent'] = self._user_agent
        # increase maximum pool size for multithreaded metadata requests
        self.session.mount('https://', requests.adapters.HTTPAdapter(pool_maxsize=16))

        self.unauth_session = requests.session()
        self.unauth_session.headers['User-Agent'] = self._user_agent

        self._oauth_basic = HTTPBasicAuth(self._user_basic, self._pw_basic)

        self.access_token = None
        self.user = None

        self.language_code = lc
        self.country_code = cc

        self.request_timeout = timeout

    def _extract_price_from_elt(self, dom_elt=None, asset_name='NO NAME') -> float:
        """
        Extracts the price from a BeautifulSoup element.
        :param dom_elt: The BeautifulSoup element.
        :param asset_name: The name of the asset (for display purpose only).
        :return: The price.
        """
        if dom_elt is None or dom_elt == '':
            self.log.debug(f'Price not found for {asset_name}')
            return 0.0
        price = 0.0
        try:
            # asset base price when logged
            price = dom_elt.text.strip('$')
            price = price.strip('€')
            price = float(price)
        except Exception as error:
            self.log.warning(f'Can not find the price for {asset_name}:{error!r}')
        return price

    def get_scrap_url(self, start=0, count=1, sort_by='effectiveDate', sort_order='DESC') -> str:
        """
        Return the scraping URL for an asset.
        """
        url = f'https://{self._url_asset_list}?start={start}&count={count}&sortBy={sort_by}&sortDir={sort_order}'
        # other possible filters
        # to see the list of possible filters: https://www.unrealengine.com/marketplace/en-US/assets and use filters on the right panel.
        """
        # can add multiple platform filters
        # Windows, Android, Linux, Mac, PS4, Nintendo%20Switch, Win32, iOS, Xbox%20One, HTML5...
        url += f'&platform={platform}'

        # can add multiple compatibleWith filters
        url += f'&compatibleWith=UE_5.2'

        # can add multiple tag filters
        url += f'&tag=22'

        url += f'&discountPercentageRange={discount_percent}'
        url += f'&priceRange=%5B{prince_min*100}%2C{prince_max*100}'
        """
        return url

    def get_owned_scrap_url(self, start=0, count=1) -> str:
        """
        Return the scraping URL for an owned asset.
        """
        # 'https://www.unrealengine.com/marketplace/api/assets/vault?start=1000&count=100'
        url = f'https://{self._url_owned_assets}?start={start}&count={count}'
        return url

    def get_marketplace_product_url(self, asset_slug: str = '') -> str:
        """
        Returns the url for the asset in the marketplace.
        :param asset_slug: The asset slug.
        :return: The url.
        """
        url = f'https://{self._url_marketplace}/en-US/product/{asset_slug}'
        return url

    def get_api_product_url(self, uid: str = '') -> str:
        """
        Returns the url for the asset using the UE API.
        :param uid: The id of the asset (not the slug, nor the catalog_id).
        :return: The url.
        """
        url = f'https://{self._url_asset}/{uid}'
        return url

    def get_scraped_asset_count(self, owned_assets_only=False) -> int:
        """
        Return the number of assets in the marketplace.
        :param owned_assets_only: Whether to only the owned assets are counted.
        """
        assets_count = 0
        if owned_assets_only:
            url = f'https://{self._url_owned_assets}'
        else:
            url = f'https://{self._url_asset_list}'
        r = self.session.get(url, timeout=self.request_timeout)
        r.raise_for_status()
        json_content = r.json()
        try:
            assets_count = json_content['data']['paging']['total']
        except Exception as error:
            self.log.warning(f'Can not get the asset count from {url}:{error!r}')
        return assets_count

    def is_valid_url(self, url='') -> bool:
        """
        Check is the url is valid (i.e. http response status is 200).
        :param url: The url to check.
        :return: True if the url is valid.
        """
        result = False
        if not url:
            return result
        try:
            r = self.session.get(url, timeout=self.request_timeout)
        except (requests.exceptions.Timeout, ConnectionError):
            self.log.warning(f'Timeout for {url}')
            return result
        if r.status_code == 200:
            result = True
        return result

    def get_json_data_from_url(self, url='') -> dict:
        """
        Return the scraped assets.
        :param url: The url to scrap.
        :return: The json data.
        """
        json_data = {}
        if not url:
            return json_data
        r = self.session.get(url, timeout=self.request_timeout)
        # r.raise_for_status() # commented line because we want the exceptions to be raised
        json_data = r.json()
        return json_data

    def resume_session(self, session: dict) -> dict:
        """
        Resumes a session.
        :param session: The session.
        :return: The session.
        """
        self.session.headers['Authorization'] = f'bearer {session["access_token"]}'
        url = f'https://{self._oauth_host}/account/api/oauth/verify'
        r = self.session.get(url, timeout=self.request_timeout)
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorMessage' in j:
            self.log.warning(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])

        # update other data
        session.update(j)
        self.user = session
        return self.user

    def start_session(
        self, refresh_token: str = None, exchange_token: str = None, authorization_code: str = None, client_credentials: bool = False
    ) -> dict:
        """
        Start a session.
        :param refresh_token: refresh token.
        :param exchange_token: exchange token.
        :param authorization_code: authorization code.
        :param client_credentials: client credentials.
        :return: The session.
        :raise: ValueError,InvalidCredentialsError.
        """
        if refresh_token:
            params = dict(grant_type='refresh_token', refresh_token=refresh_token, token_type='eg1')
        elif exchange_token:
            params = dict(grant_type='exchange_code', exchange_code=exchange_token, token_type='eg1')
        elif authorization_code:
            params = dict(grant_type='authorization_code', code=authorization_code, token_type='eg1')
        elif client_credentials:
            params = dict(grant_type='client_credentials', token_type='eg1')
        else:
            raise ValueError('At least one token type must be specified!')

        url = f'https://{self._oauth_host}/account/api/oauth/token'
        r = self.session.post(url, data=params, auth=self._oauth_basic, timeout=self.request_timeout)
        # Only raise HTTP exceptions on server errors
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorCode' in j:
            if j['errorCode'] == 'errors.com.epicgames.oauth.corrective_action_required':
                self.log.error(f'{j["errorMessage"]} ({j["correctiveAction"]}), '
                               f'open the following URL to take action: {j["continuationUrl"]}')
            else:
                self.log.error(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])
        elif r.status_code >= 400:
            self.log.error(f'EGS API responded with status {r.status_code} but no error in response: {j}')
            raise InvalidCredentialsError('Unknown error')

        self.session.headers['Authorization'] = f'bearer {j["access_token"]}'
        # only set user info when using non-anonymous login
        if not client_credentials:
            self.user = j

        return j

    def get_item_token(self) -> str:
        """
        Get the item token.
        Unused but kept for the global API reference.
        :return: The item token using json format.
        """
        url = f'https://{self._oauth_host}/account/api/oauth/exchange'
        r = self.session.get(url, timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_item_assets(self, platform='Windows', label='Live'):
        """
        Get the item assets.
        :param platform: platform to get assets for.
        :param label: label of the assets.
        :return: The item assets using json format.
        """
        url = f'https://{self._launcher_host}/launcher/api/public/assets/{platform}'
        r = self.session.get(url, params=dict(label=label), timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_item_manifest(self, namespace, catalog_item_id, app_name, platform='Windows', label='Live') -> dict:
        """
        Get the item manifest.
        :param namespace:  namespace.
        :param catalog_item_id: catalog item id.
        :param app_name: app name.
        :param platform: platform to get manifest for.
        :param label: label of the manifest.
        :return: The item manifest using json format.
        """
        url = f'https://{self._launcher_host}/launcher/api/public/assets/v2/platform/{platform}/namespace/{namespace}/catalogItem/{catalog_item_id}/app/{app_name}/label/{label}'
        r = self.session.get(url, timeout=self.request_timeout)
        r.raise_for_status()
        return r.json()

    def get_item_info(self, namespace: str, catalog_item_id: str, timeout: float = None) -> (dict, int):
        """
        Get the item info.
        :param namespace: Namespace of the item.
        :param catalog_item_id: Catalog item id of the item.
        :param timeout: Timeout for the request.
        :return: (The item info, status code).
        """
        url = f'https://{self._catalog_host}/catalog/api/shared/namespace/{namespace}/bulk/items'
        r = self.session.get(
            url,
            params=dict(
                id=catalog_item_id, includeDLCDetails=True, includeMainGameDetails=True, country=self.country_code, locale=self.language_code
            ),
            timeout=timeout or self.request_timeout
        )
        r.raise_for_status()
        return r.json().get(catalog_item_id, None), r.status_code

    def get_library_items(self, include_metadata=True) -> list:
        """
        Get the library items.
        :param include_metadata: Whether to include metadata.
        :return: The library items.
        """
        records = []
        url = f'https://{self._library_host}/library/api/public/items'
        r = self.session.get(url, params=dict(includeMetadata=include_metadata), timeout=self.request_timeout)
        r.raise_for_status()
        j = r.json()
        records.extend(j['records'])

        # Fetch remaining library entries as long as there is a cursor
        url = f'https://{self._library_host}/library/api/public/items'
        while cursor := j['responseMetadata'].get('nextCursor', None):
            r = self.session.get(url, params=dict(includeMetadata=include_metadata, cursor=cursor), timeout=self.request_timeout)
            r.raise_for_status()
            j = r.json()
            records.extend(j['records'])

        return records

    def search_for_asset_url(self, asset_name: str, timeout=10.0) -> []:
        """
        Find the asset url from the asset name by searching the asset name in the unreal engine marketplace.
        :param asset_name: asset name to search.
        :param timeout: timeout for the request.
        :return: (The asset url, the asset name (converted or found), the grab result code).
        """
        # remove the suffix _EngineVersion (ex _4.27) at the end of the name to have a valid search value
        regex = r"_[4|5]\.\d{1,2}$"
        converted_name = re.sub(regex, '', asset_name, 0)
        # Replace ' ' by '%20'
        converted_name = converted_name.replace(' ', '%20')
        # SnakeCase
        # converted_name = inflection.underscore(converted_name)
        # Lower case
        converted_name_lower = converted_name.lower()
        # Replace '_' by '-'
        # converted_name_lower = converted_name_lower.replace('_', '-')

        # remove some not alphanumeric cars (NOT ALL, keep %)
        entry_list = [':', ',', '.', ';', '=', '?', '!', '#', "/", "$", "€"]
        for entry in entry_list:
            converted_name_lower = converted_name_lower.replace(entry, '')

        url = ''
        asset_slug = converted_name_lower
        # TODO: improve the following code to use the marketplace API instead of the website using beautifulsoup
        search_url_root = f'https://{self._search_url}/assets?keywords='
        search_url_full = search_url_root + converted_name_lower
        try:
            r = self.session.get(search_url_full, timeout=timeout)
        except requests.exceptions.Timeout:
            self.log.warning(f'Timeout for {asset_name}')
            return [url, asset_slug, GrabResult.TIMEOUT.name]
        if not r.ok:
            self.log.warning(f'Can not find the url for {asset_name}:{r.reason}')
            return [url, asset_slug, GrabResult.PAGE_NOT_FOUND.name]

        soup = BeautifulSoup(r.content, 'html.parser')
        links = []
        group_elt = soup.find('div', attrs={'class': 'asset-list-group'})

        if "No content found" in group_elt.getText():
            self.log.info(f'{asset_name} has not been not found in marketplace.It has been added to the notfound_logger file')
            if self.notfound_logger:
                self.notfound_logger.info(asset_name)
            return [url, asset_slug, GrabResult.CONTENT_NOT_FOUND.name]

        # find all links to assets that correspond to the search
        for link in group_elt.findAll('a', attrs={'class': 'mock-ellipsis-item mock-ellipsis-item-helper ellipsis-text'}):
            links.append(link.get('href'))

        # return the first one (probably the best choice)
        asset_slug = links[0].replace('/marketplace/en-US/product/', '')
        url = 'https://www.unrealengine.com' + links[0]
        return [url, asset_slug, GrabResult.NO_ERROR.name]

    def grab_assets_extra(self, asset_name: str, asset_title: str, timeout=10.0, verbose_mode=False) -> dict:
        """
        Grab the extra data of an asset (price, review...) using BeautifulSoup from the marketplace.
        :param asset_name: name of the asset.
        :param asset_title: title of the asset.
        :param timeout: connection timeout.
        :param verbose_mode: verbose mode.
        :return: a dict with the extra data.
        """
        not_found_price = 0.0
        not_found_review = 0.0
        supported_versions = ''
        page_title = ''
        no_result = create_empty_assets_extra(asset_name=asset_name)

        # try to find the url of the asset by doing a search in the marketplace
        asset_url, asset_slug, error_code = self.search_for_asset_url(asset_title, timeout)

        # TODO: improve the following code to use the marketplace API instead of Scraping using beautifulsoup
        if asset_url == '' or error_code != GrabResult.NO_ERROR.name:
            self.log.info('No result found for grabbing data.\nThe asset name that has been searched for has been stored in the "Page title" Field')
            no_result['grab_result'] = error_code
            no_result['page_title'] = asset_slug
            return no_result
        try:
            response = self.session.get(asset_url)  # when using session, we are already logged in Epic game
            response.raise_for_status()
            self.log.info(f'Grabbing extra data for {asset_name}')
        except requests.exceptions.RequestException as error:
            self.log.warning(f'Can not get extra data for {asset_name}:{error!r}')
            self.log.info('No result found for grabbing data.\nThe asset name that has been searched for has been stored in the "Page title" Field')
            no_result['grab_result'] = error_code
            no_result['page_title'] = asset_slug
            return no_result

        soup_logged = BeautifulSoup(response.text, 'html.parser')

        price = not_found_price
        discount_price = not_found_price

        search_for_price = True

        # check if already owned
        owned = False
        owned_elt = soup_logged.find('div', class_='purchase')
        if owned_elt is not None:
            if 'Free' in owned_elt.getText():
                # free price when logged
                price = 0.0
                search_for_price = False
                if verbose_mode:
                    self.log.info(f'{asset_name} is free (check 1)')
            elif 'Open in Launcher' in owned_elt.getText():
                # owned asset
                owned = True
                if verbose_mode:
                    self.log.info(f'{asset_name} is already owned')
                # grab the price on a non logged soup (price will be available on that page only)
                try:
                    response = requests.get(asset_url, timeout=timeout)  # not using session, so not logged in Epic game
                    response.raise_for_status()
                    soup_not_logged = BeautifulSoup(response.text, 'html.parser')
                    owned_elt = soup_not_logged.find('div', class_='purchase')
                except requests.exceptions.RequestException:
                    pass

        if search_for_price and owned_elt is not None:
            if 'Sign in to Download' in owned_elt.getText():
                # free price when logged or not
                price = 0.0
                if verbose_mode:
                    self.log.info(f'{asset_name} is free (check 2)')
            else:
                # get price using the logged or the not logged soup
                # Note:
                #   when not discounted
                #       base-price is not available
                #       price is 'save-discount'
                #       discount-price is 0.0
                #   when discounted
                #       price is 'base-price'
                #       discount-price is 'save-discount'
                elt = owned_elt.find('span', class_='save-discount')
                current_price = self._extract_price_from_elt(elt, asset_name)
                elt = owned_elt.find('span', class_='base-price')
                base_price = self._extract_price_from_elt(elt, asset_name)
                if elt is not None:
                    # discounted
                    price = base_price
                    discount_price = current_price
                else:
                    # not discounted
                    price = current_price
                    discount_price = current_price

        # get review
        reviews_elt = soup_logged.find('div', class_='asset-detail-rating')
        if reviews_elt is not None:
            reviews_elt = reviews_elt.find('div', class_='rating-board__pop__title')
        if reviews_elt is not None:
            try:
                inner_span_elt = reviews_elt.find('span')
                content = inner_span_elt.text
                pos = content.index(' out of ')
                review = float(content[0:pos])
            except Exception as error:
                self.log.warning(f'Can not find the review for {asset_name}:{error!r}')
                review = not_found_review
        else:
            self.log.debug(f'reviews not found for {asset_name}')
            review = not_found_review

        # get Supported Engine Versions
        title_elt = soup_logged.find('span', class_='ue-version')
        if title_elt is not None:
            try:
                supported_versions = title_elt.text
            except Exception as error:
                self.log.warning(f'Can not find the Supported Engine Versions for {asset_name}:{error!r}')
        else:
            self.log.debug(f'Can not find the Supported Engine Versions for {asset_name}')
            review = not_found_review

        # get page title
        title_elt = soup_logged.find('h1', class_='post-title')
        if title_elt is not None:
            try:
                page_title = title_elt.text
            except Exception as error:
                self.log.warning(f'Can not find the Page title for {asset_name}:{error!r}')
        else:
            self.log.debug(f'Can not find the Page title not found for {asset_name}')
            review = not_found_review
        discount_percentage = 0.0 if (discount_price == 0.0 or price == 0.0 or discount_price == price) else int(
            (price - discount_price) / price * 100.0
        )
        discounted = (discount_price < price) or discount_percentage > 0.0

        self.log.info(f'GRAB results: asset_slug={asset_slug} discounted={discounted} owned={owned} price={price} review={review}')
        return {
            'asset_name': asset_name,
            'asset_slug': asset_slug,
            'price': price,
            'discount_price': discount_price,
            'review': review,
            'owned': owned,
            'discount_percentage': discount_percentage,
            'discounted': discounted,
            'asset_url': asset_url,
            'page_title': page_title,
            'supported_versions': supported_versions,
            'grab_result': error_code,
        }

    def get_asset_data_from_marketplace(self, url: str) -> dict:
        """
        Get the asset data from the marketplace using beautifulsoup.
        :param url: The url to grab.

        Note: This is the only way I know to get the id of an asset from its slug (or url)
        """
        empty_data = {
            'id': '',
            'name': '',
            'category': '',
            'description': '',
            'image': '',
            'release_date': '',
            'url': '',
            'price': '',
            'price_currency': '',
            'page_title': '',
            'grab_result': '',
        }
        json_data = empty_data.copy()
        asset_slug = url.split('/')[-1]
        json_data['page_title'] = asset_slug
        try:
            response = self.session.get(url)  # when using session, we are already logged in Epic game
            response.raise_for_status()
            self.log.info(f'Grabbing asset data from {url}')
        except requests.exceptions.RequestException as error:
            self.log.warning(f'Can not get asset data for {url}:{error!r}')
            json_data['grab_result'] = GrabResult.PAGE_NOT_FOUND.name
            return json_data

        json_data['grab_result'] = GrabResult.NO_ERROR.name

        soup = BeautifulSoup(response.content, 'html.parser')

        # Finding all script tags in the HTML
        # scripts = soup.find_all('script')
        # Finding all script tags with the type 'application/ld+json' in the HTML
        scripts = soup.find_all('script', {'type': 'application/ld+json'})

        for index, script in enumerate(scripts):
            # Remove script tags
            json_text = script.string

            # Convert to Python dictionary
            json_dict = json.loads(json_text)
            # exemple of json_dict
            # {
            #     "@context"   : "https://schema.org",
            #     "@type"      : "Product",
            #     "sku"        : "d27cf128fdc24e328cf950b019563bc5",
            #     "productID"  : "d27cf128fdc24e328cf950b019563bc5",
            #     "name"       : "Volcrate",
            #     "category"   : "Characters",
            #     "image"      : [
            #         "https://cdn1.epicgames.com/ue/item/Volcrate_FeaturedNew-894x488-ad93ea4be7589802d9dc289a4af3a751.png"
            #     ],
            #     "description": "Here is a Volcrate, this race is a crossing between a bird and a human. They mostly behave as barbarians with their impressive musculature, performing powerful devastating attacks.",
            #     "releaseDate": "2016-12-21T00:00:00.000Z",
            #     "brand"      : {
            #         "@type": "Brand",
            #         "name" : "Unreal Engine",
            #         "logo" : {
            #             "@type": "ImageObject",
            #             "url"  : "https://cdn2.unrealengine.com/Unreal+Engine%2Flogos%2FUnreal_Engine_Black-1125x1280-cfa228c80703d4ffbd1cc05eabd5ed380818da45.png"
            #         }
            #     },
            #     "offers"     : {
            #         "@type"        : "Offer",
            #         "price"        : "€32.01",
            #         "availability" : "http://schema.org/InStock",
            #         "priceCurrency": "EUR",
            #         "url"          : "https://www.unrealengine.com/marketplace/en-US/product/volcrate"
            #     }
            # }

            # check if the script describes a product
            if json_dict['@type'] == 'Product':
                try:
                    json_data['id'] = json_dict['productID']
                    json_data['name'] = json_dict['name']
                    json_data['category'] = json_dict['category']
                    json_data['description'] = json_dict['description']
                    json_data['image'] = json_dict['image']
                    json_data['release_date'] = json_dict['releaseDate']
                    json_data['url'] = json_dict['offers']['url']  # must be the same as url
                    json_data['price'] = json_dict['offers']['price']
                    json_data['price_currency'] = json_dict['offers']['priceCurrency']
                except KeyError as error:
                    self.log.warning(f"A key is missing in Script {index + 1}: {error!r}")
                    json_data['grab_result'] = GrabResult.PARTIAL.name
                    continue
        if json_data['url'].lower().replace('www.', '') != url.lower().replace('www.', ''):
            self.log.warning(f"URLs do not match: {json_data['url']} != {url}")
            json_data['grab_result'] = GrabResult.INCONSISTANT_DATA.name
        return json_data



# coding=utf-8
"""
Windows specific helper functions for registry access.
"""
import ctypes


def double_clicked() -> bool:
    """
    Check if the application was started by a double click.
    """
    # Thanks https://stackoverflow.com/a/55476145

    # Load kernel32.dll
    kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
    # Create an array to store the processes in.  This doesn't actually need to
    # be large enough to store the whole process list since GetConsoleProcessList()
    # just returns the number of processes if the array is too small.
    process_array = (ctypes.c_uint * 1)()
    num_processes = kernel32.GetConsoleProcessList(process_array, 1)
    return num_processes < 3



# coding=utf-8
"""
Definition for the types used in this module:
- UEAssetType: enum to represent the asset type
- DataSourceType: an enum to represent the data source type
- WidgetType: enum for the widget types.
"""
from enum import Enum


class DataFrameUsed(Enum):
    """ Enum to represent the data frame used for getting or setting the data"""
    AUTO = 0  # Automatically select the data frame wether it is filtered or not
    UNFILTERED = 1  # Use the data frame with unfiltered data
    FILTERED = 2  # Use the data frame with filtered data
    MODEL = 3  # Use the model.df.
    BOTH = 4  # Use both data frames


class UEAssetType(Enum):
    """ Enum to represent the asset type """
    Unknown = 0
    Plugin = 1
    Asset = 2
    Manifest = 3

    @property
    def category_name(self):
        """ Return the category name of the asset type """
        if self == self.Plugin:
            return 'local/plugins'  # existing category in the marketplace
        if self == self.Asset:
            return 'local/asset'  # non-existing category in the marketplace
        if self == self.Manifest:
            return 'local/manifest'  # non-existing category in the marketplace
        return 'local/unknown'  # non-existing category in the marketplace


class DataSourceType(Enum):
    """ Enum to represent the data source type """
    FILE = 1
    SQLITE = 2


class WidgetType(Enum):
    """ Enum for the widget types """
    ENTRY = 0  # Entry widget
    TEXT = 1  # Text widget
    LABEL = 2  # Label widget
    CHECKBUTTON = 3  # Checkbutton widget
    BUTTON = 4  # Button widget



# coding=utf-8
"""
implementation for:
- UEAssetDbHandler: Handles database operations for the UE Assets.
"""
import csv
import datetime
import inspect
import logging
import os
import random
import sqlite3
from pathlib import Path

from faker import Faker

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.core import default_datetime_format
from UEVaultManager.lfs.utils import path_join
from UEVaultManager.models.csv_sql_fields import CSVFieldState, get_sql_field_name_list
from UEVaultManager.models.types import DbVersionNum
from UEVaultManager.models.UEAssetClass import UEAsset
from UEVaultManager.tkgui.modules.functions import create_file_backup, update_loggers_level
from UEVaultManager.tkgui.modules.functions_no_deps import convert_to_str_datetime, create_uid, path_from_relative_to_absolute
from UEVaultManager.tkgui.modules.types import UEAssetType
from UEVaultManager.utils.cli import check_and_create_path


class UEAssetDbHandler:
    """
    Handles database operations for the UE Assets.
    :param database_name: The name of the database file.
    :param reset_database: Whether the database will be reset.

    Note: The database will be created if it doesn't exist.
    """
    logger = logging.getLogger(__name__.split('.')[-1])  # keep only the class name
    update_loggers_level(logger)
    db_version: DbVersionNum = DbVersionNum.V0  # updated in check_and_upgrade_database()

    def __init__(self, database_name: str, reset_database: bool = False):
        self.connection = None
        # the user fields that must be preserved when updating the database
        # these fields are also present in the asset table and in the UEAsset.init_data() method
        # THEY WILL BE PRESERVED when parsing the asset data
        self.user_fields = get_sql_field_name_list(filter_on_states=[CSVFieldState.USER])
        # the field we keep for previous data. NEED TO BE SEPARATED FROM self.user_fields
        # THEY WILL BE USED (BUT NOT FORCELY PRESERVED) when parsing the asset data
        self.preserved_data_fields = self.user_fields
        self.preserved_data_fields.append('id')
        changed_fields = get_sql_field_name_list(filter_on_states=[CSVFieldState.CHANGED])
        self.preserved_data_fields.extend(changed_fields)
        # we also need to preserve the values in the database that are not in the (CSV) table
        asset_fields = get_sql_field_name_list(filter_on_states=[CSVFieldState.ASSET_ONLY])
        self.preserved_data_fields.extend(asset_fields)
        self.database_name: str = database_name
        self._init_connection()
        if reset_database:
            self.drop_tables()

        self.check_and_upgrade_database()

    class DatabaseConnection:
        """
        Context manager for opening and closing a database connection.
        :param database_name: The name of the database file.
        """

        def __init__(self, database_name: str):
            self.database_name: str = database_name
            self.conn = sqlite3.connect(self.database_name, check_same_thread=False)

        def __enter__(self):
            """
            Open the database connection.
            :return: The sqlite3.Connection object.
            """
            self.conn = sqlite3.connect(self.database_name, check_same_thread=False)
            return self.conn

        def __exit__(self, exc_type, exc_val, exc_tb):
            """
            Close the database connection.
            :param exc_type: Exception type.
            :param exc_val: Exception value.
            :param exc_tb: Exception traceback.
            """
            self.conn.close()
            self.conn = None

    def __del__(self):
        # log could alrready be destroyed before here
        # self.logger.debug('Deleting UEAssetDbHandler and closing connection')
        print('Deleting UEAssetDbHandler and closing connection')
        self._close_connection()

    def _close_connection(self) -> None:
        """
        Close the database connection.
        """
        if self.connection is not None:
            try:
                self.connection.close()
            except sqlite3.Error as error:
                print(f'Error while closing sqlite connection: {error!r}')
        self.connection = None

    def _init_connection(self) -> sqlite3.Connection:
        """
        Initialize the database connection.
        :return: The sqlite3.Connection object.

        Note: It will also set self.Connection property.
        """
        self._close_connection()  # close the connection IF IT WAS ALREADY OPENED
        try:
            self.connection = self.DatabaseConnection(self.database_name).conn
        except sqlite3.Error as error:
            print(f'Error while connecting to sqlite: {error!r}')

        return self.connection

    def _get_db_version(self) -> DbVersionNum:
        """
        Check the database version.
        :return: The database version.
        """
        try:
            if self.connection is not None:
                cursor = self.connection.cursor()
                cursor.execute("PRAGMA user_version")
                ret = int(cursor.fetchone()[0])  # type: int
                cursor.close()
                db_version = DbVersionNum(ret)
                return db_version
        except sqlite3.OperationalError:
            self.logger.critical('Could not get the database version')
            return DbVersionNum.V0

    def _set_db_version(self, new_version: DbVersionNum) -> None:
        """
        Set the database version.
        :param new_version: The new database version.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f'PRAGMA user_version = {new_version.value}')
            cursor.close()
            self.logger.info(f'database version is now set to {new_version}')

    def _check_db_version(self, minimal_db_version: DbVersionNum, caller_name='this method') -> bool:
        """
        Check if the database version is compatible with the current method.
        :param minimal_db_version: The minimal database version that is compatible with the current method.
        :param caller_name: The name of the method that called this method.
        :return:  True if the database version is compatible, otherwise False.
        """
        if self.db_version.value < minimal_db_version.value:
            message = f'This version of {caller_name} is only compatible with {minimal_db_version} or more database.\nPlease upgrade the database first.'
            self.logger.critical(message)
            return False
        return True

    def _add_missing_columns(self, table_name: str, required_columns: dict) -> None:
        """
        Add missing columns to a table.
        :param table_name: Name of the table.
        :param required_columns: Dictionary of columns to add. Key is the column name, value is the data type.

        Note: the AFTER parameter in SQL is not supported in the SQLite version used.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"PRAGMA table_info({table_name})")
            columns = {row[1]: row for row in cursor.fetchall()}
            for column_name, data_type in required_columns.items():
                if column_name not in columns:
                    query = f"ALTER TABLE {table_name} ADD COLUMN {column_name} {data_type}"
                    cursor.execute(query)
            self.connection.commit()
            cursor.close()

    def _run_query(self, query: str, data: dict = None) -> list:
        """
        Run a query.
        :param query: The query to run.
        :param data: A dictionary containing the data to use in the query.
        :return: The result of the query.
        """
        result = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            if data is None:
                cursor.execute(query)
            else:
                cursor.execute(query, data)
            self.connection.commit()
            result = cursor.fetchall()
            cursor.close()
        return result

    def _insert_or_update_row(self, table_name: str, row_data: dict) -> bool:
        """
        Insert or update a row in the given table.
        :param table_name: The name of the table.
        :param row_data: A dictionary representing the row to insert or update.
        :return: True if the row was inserted or updated, otherwise False.
        """
        if not table_name or not row_data:
            return False
        uid = row_data.get('id', None)  # check if the row as an id to check
        # remove all fields whith a None Value
        filtered_fields = {k: v for k, v in row_data.items() if (v is not None and v not in ('nan', 'None'))}
        if len(filtered_fields) == 0:
            return False
        column_list = filtered_fields.keys()
        cursor = self.connection.cursor()
        if uid is not None:
            # check if the asset already exists in the database
            query = "SELECT id FROM " + table_name + " WHERE id = ?"  # we don't use a {table_name} here to avoid inspection warning
            cursor.execute(query, (uid, ))
            result = cursor.fetchone()
        else:
            result = None
        if result is None:
            # uid is not in row fields, or the row with uid does not exist in the database, add it
            fields = ", ".join(f"{column}" for column in column_list)
            values = ", ".join(f":{column}" for column in column_list)
            # this query will insert or update the asset if it already exists.
            # noinspection SqlInsertValues
            query = f"REPLACE INTO {table_name} ({fields}) VALUES ({values})"
        else:
            # the row with uid already exists in the database, update it
            fields = ", ".join(f"{column} = :{column}" for column in column_list)
            query = "UPDATE " + table_name + f" SET {fields} WHERE id = '{uid}'"  # we don't use a {table_name} here to avoid inspection warning
        try:
            cursor.execute(query, row_data)
            return True
        except (sqlite3.IntegrityError, sqlite3.InterfaceError) as error:
            self.logger.warning(f"Error while inserting/updating row with id '{uid}': {error!r}")
            return False

    def db_exists(self) -> bool:
        """
        Check if the database file exists.
        :return: True if the database file exists, otherwise False.
        """
        return os.path.exists(self.database_name)

    def create_tables(self, upgrade_to_version=DbVersionNum.V1) -> None:
        """
        Create the tables if they don't exist.
        :param upgrade_to_version: The database version we want to upgrade TO.
        """
        # all the following steps must be run sequentially
        if upgrade_to_version.value >= DbVersionNum.V1.value:
            if self.connection is not None:
                cursor = self.connection.cursor()
                # create the first version of the database
                # Note:
                # - this table has the same structure as the json files saved inside the method UEAssetScraper.save_to_file()
                # - the order of columns must match the order of the fields in UEAsset.init_data() method.
                query = """
                CREATE TABLE IF NOT EXISTS assets ( 
                    id TEXT PRIMARY KEY NOT NULL, 
                    namespace TEXT, 
                    catalog_item_id TEXT, 
                    title TEXT, 
                    category TEXT, 
                    author TEXT, 
                    thumbnail_url TEXT, 
                    asset_slug TEXT, 
                    currency_code TEXT, 
                    description TEXT, 
                    technical_details TEXT, 
                    long_description TEXT, 
                    tags TEXT, 
                    comment_rating_id TEXT, 
                    rating_id TEXT, 
                    status TEXT, 
                    price REAL, 
                    discount_price REAL, 
                    discount_percentage REAL, 
                    is_catalog_item BOOLEAN, 
                    is_new BOOLEAN, 
                    free BOOLEAN, 
                    discounted BOOLEAN, 
                    can_purchase BOOLEAN, 
                    owned INTEGER, 
                    review REAL, 
                    review_count INTEGER,
                    custom_attributes TEXT
                )
                """
                cursor.execute(query)
                self.connection.commit()
                cursor.close()
        if upgrade_to_version.value >= DbVersionNum.V3.value:
            cursor = self.connection.cursor()
            query = "CREATE TABLE IF NOT EXISTS last_run (date DATETIME, mode TEXT, files_count INTEGER, items_count INTEGER, scraped_ids TEXT)"
            cursor.execute(query)
            self.connection.commit()
            cursor.close()
        if upgrade_to_version.value >= DbVersionNum.V7.value:
            cursor = self.connection.cursor()
            query = "CREATE TABLE IF NOT EXISTS tags (id INTEGER PRIMARY KEY NOT NULL, name TEXT)"
            cursor.execute(query)
            self.connection.commit()
            cursor.close()
        if upgrade_to_version.value >= DbVersionNum.V8.value:
            cursor = self.connection.cursor()
            query = "CREATE TABLE IF NOT EXISTS ratings (id TEXT PRIMARY KEY NOT NULL, averageRating REAL, total INTEGER)"
            cursor.execute(query)
            self.connection.commit()
            cursor.close()
        if upgrade_to_version.value >= DbVersionNum.V9.value:
            cursor = self.connection.cursor()
            query = """
                    CREATE VIEW IF NOT EXISTS assets_tags AS
                    WITH RECURSIVE split(asset_id, tags, tag, rest) AS (
                        SELECT asset_id, tags, '', tags || ','
                        FROM assets
                        UNION ALL
                        SELECT asset_id, tags, substr(rest, 0, instr(rest, ',')), substr(rest, instr(rest, ',')+1)
                        FROM split
                        WHERE rest <> ''
                    )
                    SELECT asset_id, tag FROM split WHERE tag <> '';
                    """
            cursor.execute(query)
            self.connection.commit()
            cursor.close()

    def check_and_upgrade_database(self, upgrade_from_version: DbVersionNum = None) -> None:
        """
        Change the tables structure according to different versions.
        :param upgrade_from_version: The version we want to upgrade FROM. if None, the current version will be used.
        """
        if not self.is_table_exist('assets'):
            previous_version = DbVersionNum.V0
        elif not self.is_table_exist('tags'):
            previous_version = DbVersionNum.V6
        elif not self.is_table_exist('ratings'):
            previous_version = DbVersionNum.V7
        else:
            # force an update of the db_version property
            previous_version = self._get_db_version()
            # if self.db_version == DbVersionNum.V0:
            #    raise Exception("Invalid database version or database is badly initialized")

        self.db_version = previous_version

        if upgrade_from_version is None:
            upgrade_from_version = self.db_version

        # all the following steps must be run sequentially
        if upgrade_from_version.value <= DbVersionNum.V1.value:
            self.db_version = upgrade_from_version = DbVersionNum.V2
            self.create_tables(upgrade_to_version=self.db_version)
        if upgrade_from_version == DbVersionNum.V2:
            self.db_version = upgrade_from_version = DbVersionNum.V3
            self.create_tables(upgrade_to_version=self.db_version)
        if upgrade_from_version == DbVersionNum.V3:
            self._add_missing_columns(
                'assets',
                required_columns={
                    'asset_id': 'TEXT',
                    'asset_url': 'TEXT',
                    'comment': 'TEXT',
                    'stars': 'INTEGER',
                    'must_buy': 'BOOLEAN',
                    'test_result': 'TEXT',
                    'installed_folder': 'TEXT',
                    'alternative': 'TEXT',
                    'origin': 'TEXT',
                    'added_manually': 'BOOLEAN'
                }
            )
            self.db_version = upgrade_from_version = DbVersionNum.V4
        if upgrade_from_version == DbVersionNum.V4:
            self._add_missing_columns(
                'assets',
                required_columns={
                    'added_manually': 'BOOLEAN',
                    'page_title': 'TEXT',
                    'obsolete': 'BOOLEAN',
                    'supported_versions': 'TEXT',
                    'creation_date': 'DATETIME',
                    'update_date': 'DATETIME',
                    'date_added_in_db': 'DATETIME',
                    'grab_result': 'TEXT',
                    'old_price': 'REAL'
                }
            )
            self.db_version = upgrade_from_version = DbVersionNum.V5
        if upgrade_from_version == DbVersionNum.V5:
            self._add_missing_columns('assets', required_columns={'tags': 'TEXT', })
            self.db_version = upgrade_from_version = DbVersionNum.V6
        if upgrade_from_version == DbVersionNum.V6:
            self.db_version = upgrade_from_version = DbVersionNum.V7
            self.create_tables(upgrade_to_version=self.db_version)
        if upgrade_from_version == DbVersionNum.V7:
            self.db_version = upgrade_from_version = DbVersionNum.V8
            self.create_tables(upgrade_to_version=self.db_version)
        if upgrade_from_version == DbVersionNum.V8:
            self.db_version = upgrade_from_version = DbVersionNum.V9
            self.create_tables(upgrade_to_version=self.db_version)
        if upgrade_from_version == DbVersionNum.V9:
            # next line produce error: sqlite3.OperationalError: Cannot add a PRIMARY KEY column
            # self._add_missing_columns('last_run', required_columns={'id': 'INTEGER PRIMARY KEY AUTOINCREMENT'})
            query = "DROP TABLE last_run"
            self._run_query(query)
            query = """
            CREATE TABLE last_run (
                date        DATETIME,
                mode        TEXT,
                files_count INTEGER,
                items_count INTEGER,
                scraped_ids TEXT,
                id          INTEGER PRIMARY KEY AUTOINCREMENT
            );"""
            self._run_query(query)
            self.db_version = upgrade_from_version = DbVersionNum.V10
        if upgrade_from_version == DbVersionNum.V10:
            """
            does not exist yet
            """
            """
            # do some stuff here
            self.db_version = upgrade_from_version = DbVersionNum.V11
            """
            pass
        if previous_version != self.db_version:
            self.logger.info(f'Database upgraded to {self.db_version}')
            self._set_db_version(self.db_version)

    def is_table_exist(self, table_name) -> bool:
        """
        Check if a table exists.
        :param table_name: The name of the table to check.
        :return:  True if the 'assets' table exists, otherwise False.
        """
        result = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'")
            result = cursor.fetchone()
            cursor.close()
        return result is not None

    def get_rows_count(self, table_name='assets') -> int:
        """
        Get the number of rows in the given table.
        :param table_name: The name of the table.
        :return:  The number of rows in the 'assets' table.
        """
        row_count = 0
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
            row_count = cursor.fetchone()[0]
            cursor.close()
        return row_count

    # noinspection DuplicatedCode
    def save_last_run(self, data: dict):
        """
        Save the last run data into the 'last_run' table.
        :param data: A dictionary containing the data to save.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(DbVersionNum.V3, caller_name=inspect.currentframe().f_code.co_name):
            return
        if self.connection is not None:
            cursor = self.connection.cursor()
            query = "INSERT INTO last_run (date, mode, files_count, items_count, scraped_ids) VALUES (:date, :mode, :files_count, :items_count, :scraped_ids)"
            cursor.execute(query, data)
            cursor.close()
            self.connection.commit()

    def set_assets(self, assets) -> None:
        """
        Insert or update assets into the 'assets' table.
        :param assets: A dictionary or a list of dictionaries representing assets.

        Note: the (existing) user fields data should have already been added or merged the asset dictionary.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(DbVersionNum.V2, caller_name=inspect.currentframe().f_code.co_name):
            return
        if self.connection is not None:
            if not isinstance(assets, list):
                assets = [assets]
            str_today = datetime.datetime.now().strftime(default_datetime_format)
            # Note: the order of columns and value must match the order of the fields in UEAsset.init_data() method
            for asset in assets:
                asset['update_date'] = str_today
                asset['creation_date'] = convert_to_str_datetime(value=asset['creation_date'], date_format=default_datetime_format)
                asset['date_added_in_db'] = convert_to_str_datetime(
                    value=asset['date_added_in_db'], date_format=default_datetime_format, default=str_today
                )
                tags = asset.get('tags', [])
                tags_str = self.convert_tag_list_to_string(tags)
                asset['tags'] = tags_str
                self._insert_or_update_row('assets', asset)
        try:
            self.connection.commit()
        except (sqlite3.IntegrityError, sqlite3.InterfaceError) as error:
            self.logger.warning(f'Error when committing the database changes: {error!r}')

    def get_assets_data(self, fields='*', uid=None) -> dict:
        """
        Get data from all the assets in the 'assets' table.
        :param fields: list of fields to return.
        :param uid: The id of the asset to get data for. If None, all the assets will be returned.
        :return: dictionary {ids, rows}.
        """
        if not isinstance(fields, str):
            fields = ', '.join(fields)
        row_data = {}
        where_clause = f"WHERE id='{uid}'" if id is not None else ''
        if self.connection is not None:
            self.connection.row_factory = sqlite3.Row
            cursor = self.connection.cursor()
            cursor.execute(f"SELECT {fields} FROM assets {where_clause}")
            for row in cursor.fetchall():
                uid = row['id']
                row_data[uid] = dict(row)
            cursor.close()
        return row_data

    def get_assets_data_for_csv(self, where_clause='') -> list:
        """
        Get data from all the assets in the 'assets' table for a "CSV file" like format.
        :param where_clause: A string containing the WHERE clause to use in the SQL query.
        :return: list(rows).
        """
        rows = []
        if self.connection is not None:
            cursor = self.connection.cursor()
            # generate column names for the CSV file using AS to rename the columns
            fields = get_sql_field_name_list(exclude_csv_only=True, return_as_string=True, add_alias=True)
            query = f"SELECT {fields} FROM assets"
            if where_clause:
                query += f" WHERE {where_clause}"
            if not gui_g.s.assets_order_col:
                gui_g.s.assets_order_col = 'date_added_in_db'
            query += f' ORDER by {gui_g.s.assets_order_col} DESC'
            if gui_g.s.testing_switch == 1:
                query += ' LIMIT 3000'
            try:
                cursor.execute(query)
                rows = cursor.fetchall()
                cursor.close()
            except sqlite3.OperationalError as error:
                self.logger.warning(f"Error while getting assets data: {error!r}")
        return rows

    def get_columns_name_for_csv(self) -> list:
        """
        Get the columns name from the 'assets' table in a "CSV file" like format.
        :return: list (column_names).
        """
        csv_column_names = []
        if self.connection is not None:
            cursor = self.connection.cursor()
            # generate column names for the CSV file using AS to rename the columns
            fields = get_sql_field_name_list(exclude_csv_only=True, return_as_string=True, add_alias=True)
            query = f"SELECT {fields} FROM assets ORDER BY date_added_in_db DESC LIMIT 1"
            try:
                cursor.execute(query)
                csv_column_names = [
                    description[0] for description in cursor.description
                ]  # by using the 'AS' in the SQL query, the column names are the CSV column names
                cursor.close()
            except sqlite3.OperationalError as error:
                self.logger.warning(f"Error while getting columns name: {error!r}")
        return csv_column_names

    def create_empty_row(self, return_as_string=True, empty_cell='None', empty_row_prefix='dummy_row_', do_not_save=False):
        """
        Create an empty row in the 'assets' table.
        :param return_as_string: True to return the row as a string, False to.
        :param empty_cell: The value to use for empty cells.
        :param empty_row_prefix: The prefix to use for the row ID.
        :param do_not_save: True to not save the row in the database.
        :return: A row (dict) or a string representing the empty row.
        """
        result = '' if return_as_string else {}
        # add a new row to the 'assets' table
        if self.connection is not None:
            cursor = self.connection.cursor()
            # generate a unique ID and check if it already exists in the database
            uid = ''
            count = 1
            while count > 0:
                uid = create_uid()
                cursor.execute("SELECT COUNT(*) FROM assets WHERE id = ?", (uid, ))
                count = cursor.fetchone()[0]
            cursor.close()
            ue_asset = UEAsset()
            ue_asset.data['asset_id'] = empty_row_prefix + uid  # dummy unique Asset_id to avoid issue
            ue_asset.data['thumbnail_url'] = empty_cell  # avoid displaying image warning on mouse over
            ue_asset.data['added_manually'] = True
            ue_asset.data['id'] = uid
            ue_asset.data['category'] = UEAssetType.Unknown.category_name
            if not do_not_save:
                self.save_ue_asset(ue_asset)
            if return_as_string:
                result = str(ue_asset)
            else:
                # we read the new row from the database to get CSV column names
                result = self.get_assets_data_for_csv(where_clause=f"id = '{uid}'")
        return result

    def read_ue_asset(self, uid: str) -> UEAsset:
        """
        Read an UEAsset object from the 'assets' table.
        :param uid: The ID of the asset to get.
        :return: UEAsset object.
        """
        ue_asset = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("SELECT * assets WHERE id = ?", (uid, ))
            row = cursor.fetchone()
            cursor.close()
            ue_asset = UEAsset()
            ue_asset.init_from_list(data=row)
        return ue_asset

    def save_ue_asset(self, ue_asset: UEAsset) -> None:
        """
        Save an UEAsset object to the 'assets' table.
        :param ue_asset: UEAsset object to save.
        """
        self.set_assets([ue_asset.data])

    def delete_asset(self, uid: str = '', asset_id: str = '') -> None:
        """
        Delete an asset from the 'assets' table by its ID or asset_id.
        :param uid: The ID of the asset to delete.
        :param asset_id: The Asset_id of the asset to delete. If both uid and asset_id are provided, only asset_id is used.
        """
        if self.connection is not None and (uid or asset_id):
            cursor = self.connection.cursor()
            if not asset_id:
                cursor.execute("DELETE FROM assets WHERE id = ?", (uid, ))
            else:
                cursor.execute("DELETE FROM assets WHERE asset_id = ?", (asset_id, ))
            self.connection.commit()
            cursor.close()

    def delete_all_assets(self, keep_added_manually=True) -> None:
        """
        Delete all assets from the 'assets' table.
        :param keep_added_manually: True to keep the assets added manually, False to delete all assets.
        """
        if keep_added_manually:
            where_clause = "added_manually = 0"
        else:
            where_clause = "1"
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("DELETE FROM assets WHERE ?", (where_clause, ))
            self.connection.commit()
            cursor.close()

    def update_asset(self, column: str, value, uid: str = '', asset_id: str = '') -> None:
        """
        Update a specific column of an asset in the 'assets' table by its ID.
        :param column: The name of the column.
        :param value: The new value.
        :param uid: The ID of the asset to delete.
        :param asset_id: The Asset_id of the asset to delete. If both uid and asset_id are provided, only asset_id is used.
        """
        if self.connection is not None and (uid or asset_id):
            cursor = self.connection.cursor()
            if not asset_id:
                cursor.execute(f"UPDATE assets SET {column} = ? WHERE id = ?", (value, uid))
            else:
                cursor.execute(f"UPDATE assets SET {column} = ? WHERE asset_id = ?", (value, asset_id))
            self.connection.commit()
            cursor.close()

    # noinspection DuplicatedCode
    def save_tag(self, data: dict):
        """
        Save a tag into the 'tag' table.
        :param data: A dictionary containing the data to save.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(DbVersionNum.V7, caller_name=inspect.currentframe().f_code.co_name):
            return
        if data.get('id', None) is None or data.get('name', None) is None:
            return
        if self.get_tag_by_id(data['id']) is None:
            cursor = self.connection.cursor()
            query = "INSERT INTO tags (id, name) VALUES (:id, :name)"
            cursor.execute(query, data)
            self.connection.commit()
            cursor.close()

    def get_tag_by_id(self, uid: int) -> str:
        """
        Read a tag name using its id from the 'tags' table.
        :param uid: The ID of the tag to get.
        :return: name of the tag.
        """
        result = None
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("SELECT name from tags WHERE id = ?", (uid, ))
            row = cursor.fetchone()
            cursor.close()
            result = row['name'] if row else result
        return result

    # noinspection DuplicatedCode
    def save_rating(self, data: dict):
        """
        Save a tag into the 'rating' table.
        :param data: A dictionary containing the data to save.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(DbVersionNum.V8, caller_name=inspect.currentframe().f_code.co_name):
            return
        if data.get('id', None) is None or data.get('averageRating', None) is None:
            return
        if self.get_rating_by_id(data['id']) is None:
            cursor = self.connection.cursor()
            query = "INSERT INTO ratings (id, averageRating, total) VALUES (:id, :averageRating, :total)"
            cursor.execute(query, data)
            self.connection.commit()
            cursor.close()

    def get_rating_by_id(self, uid: int) -> ():
        """
        Read a rating using its id from the 'ratings' table.
        :param uid: The ID of the ratings to get.
        :return: tuple (averageRating, total)
        """
        result = (None, None)
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("SELECT averageRating, total from ratings WHERE id = ?", (uid, ))
            row = cursor.fetchone()
            cursor.close()
            result = (row['averageRating'], row['total']) if row else result
        return result

    def get_rows_with_tags_to_convert(self, tag_value: int = None) -> list:
        """
        Get all rows from the 'assets_tags' table that could be scrapped to fix their tags.
        :param tag_value: If not None, only get the rows that have a tag == tag_value.
        :return: A list of asset_id.

        Note: if tag_value is None, the returned list contains the asset_id that have at least a tag in the tags field that is an id in the tag table
        """
        rows = []
        if self.connection is not None:
            cursor = self.connection.cursor()
            if tag_value is None:
                # get the asset_id that have at least a tag in the tags field that is an id in the tag table
                cursor.execute("SELECT DISTINCT asset_id FROM assets_tags WHERE tag GLOB '*[^0-9]*' = 0 AND tag IN (SELECT id FROM tags)")
            else:
                # get the asset_id that have at least a tag in the tags field that is id == tag_value
                cursor.execute("SELECT asset_id FROM assets_tags WHERE CAST(tag AS INTEGER) == ?", tag_value)
            rows = cursor.fetchall()
            cursor.close()
        return [row[0] for row in rows]

    def convert_tag_list_to_string(self, tags: None) -> str:
        """
        Convert a tags id list to a comma separated string of tag names.
        """
        tags_str = ''
        prefix = gui_g.s.tag_prefix  # prefix to add to the tag that has been checked
        if tags is not None and tags != [] and tags != {} and tags != '':
            if isinstance(tags, str):
                # noinspection PyBroadException
                try:
                    tags = tags.split(',')  # convert the string to a list
                except Exception:
                    return tags_str
            if isinstance(tags, list):
                names = []
                for item in tags:
                    # remove the prefix if it exists to check if these id has a name in the tag table since the last check
                    if isinstance(item, str) and item.startswith(prefix):
                        # the tag (number) as already be checked but its name was not in the database yet
                        item = item[len(prefix):]
                        try:
                            item = int(item)
                        except ValueError:
                            pass
                    if isinstance(item, int):
                        # the tag is a number (old version), whe check if it's in the database and get its name
                        name = self.get_tag_by_id(uid=item)
                        if name is None:
                            name = prefix + str(item)  # we add a suffix to the tag that has been checked
                    elif isinstance(item, dict):
                        # the tag is a dict (new version), we get its name and add it to the database if it's not already there
                        uid = item.get('id', None)  # not used for now
                        name = item.get('name', '').title()
                        self.save_tag({'id': uid, 'name': name})
                    else:
                        name = str(item).title()  # convert to string and capitalize
                    if name and name not in names:
                        names.append(name)
                tags_str = ','.join(names)
        return tags_str

    def drop_tables(self) -> None:
        """
        Drop the 'assets' and 'last_run' tables.
        """
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("DROP VIEW IF EXISTS assets_tags")
            cursor.execute("DROP TABLE IF EXISTS tags")
            cursor.execute("DROP TABLE IF EXISTS last_run")
            cursor.execute("DROP TABLE IF EXISTS assets")
            self.connection.commit()
            cursor.close()

    def get_table_names(self) -> list:
        """
        Get the names of all the tables in the database.
        :return: A list of table names.
        """
        result = []
        if self.connection is not None:
            cursor = self.connection.cursor()
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            result = [name[0] for name in cursor.fetchall()]
            cursor.close()
        return result

    def export_to_csv(
        self,
        folder_for_csv_files: str,
        table_name: str = '',
        fields: str = '',
        backup_existing=False,
        suffix_separator: str = '_##',
        suffix: str = ''
    ) -> [str]:
        """
        Export the database to a CSV file.
        :param folder_for_csv_files: The folder where the CSV files will be saved.
        :param table_name: The name of the table to export. If None, all the tables will be exported.
        :param fields: The fields to export. If None, all the fields will be exported.
        :param backup_existing: True to back up the existing CSV file before overwriting it.
        :param suffix_separator: The separator used to separate the table name from the suffix in the CSV file name.
        :param suffix: A suffix to add to the CSV file name.
        :return: list of CSV files that have been writen, [] if none.

        Note: each table will be exported to a separate CSV file using the table name as the file name.
        """
        result = []
        if self.connection is not None:
            folder_for_csv_files_p = Path(folder_for_csv_files)
            if not folder_for_csv_files_p.exists():
                folder_for_csv_files_p.mkdir(parents=True)
                self.logger.info(f'Folder "{folder_for_csv_files_p}" has been created.')
            cursor = self.connection.cursor()
            if not fields:
                fields = '*'
            if table_name:
                table_names = [table_name]
            else:
                table_names = self.get_table_names()
            for table_name in table_names:
                suffix_all = suffix_separator + suffix if suffix and suffix_separator else ''
                file_name_p = folder_for_csv_files_p / f'{table_name}{suffix_all}.csv'
                file_name = str(file_name_p)
                if backup_existing:
                    create_file_backup(file_src=file_name, path=folder_for_csv_files)
                # Get column names
                if fields == '*':
                    cursor.execute(f"PRAGMA table_info({table_name});")
                    column_names = [column[1] for column in cursor.fetchall()]
                else:
                    column_names = fields.split(',')
                query = f"SELECT {fields} FROM {table_name}"
                rows = cursor.execute(query).fetchall()
                try:
                    with open(file_name_p, 'w', newline='', encoding='utf-8') as f:
                        writer = csv.writer(f, dialect='unix')
                        # Write column names
                        writer.writerow(column_names)
                        # Write rows
                        writer.writerows(rows)
                        result.append(f'Export: {file_name}')
                except Exception as error:
                    msg = f'Error while exporting table "{table_name}" to CSV file "{file_name}"'
                    result.append(msg)
                    self.logger.warning(f'{msg}: {error!r}')
            cursor.close()
        return result

    def import_from_csv(
        self,
        folder_for_csv_files: str,
        table_name: str = '',
        delete_content: bool = False,
        is_partial: bool = False,
        suffix_separator: str = '_##',
        suffix_to_ignore=None
    ) -> ([str], bool):
        """
        Import the database from a CSV file.
        :param folder_for_csv_files: The folder containing the CSV files to import.
        :param table_name: The name of the table to import. If None, all the tables will be imported.
        :param delete_content: True to delete the content of the table before importing the data, False to append the data to the table.
        :param is_partial: True to indicate the import is partial. It True, only some fields are imported and the columns will not be checked.
        :param suffix_separator: The separator used to separate the table name from the suffix in the CSV file name.
        :param suffix_to_ignore: A list of suffix to ignore when importing the CSV files.
        :return: (list of CSV files that have been read, True if the database must be reloaded).
        """
        if suffix_to_ignore is None:
            suffix_to_ignore = []
        result = []
        must_reload = False
        if self.connection is not None:
            folder_for_csv_files = Path(folder_for_csv_files)
            if not folder_for_csv_files.exists():
                self.logger.error(f'Folder "{folder_for_csv_files}" does not exist.')
                return False
            cursor = self.connection.cursor()
            table_is_done = []  # used to avoid importing the same table multiple times
            csv_files = list(folder_for_csv_files.glob('*.csv'))
            # sort the list in reverse order, as it when the same table has multiple files to import from, we import the last one (with datetime suffix) first
            # csv_files = sorted(csv_files, reverse=True)
            given_table_name = table_name
            for file_name_p in csv_files:
                success_count = 0
                fails_count = 0
                file_name = str(file_name_p)

                # check if the file is not empty
                if not file_name_p.exists() or file_name_p.stat().st_size == 0:
                    msg = f'The CSV file "{file_name}" does not exist or is empty and will not be imported.'
                    result.append(msg)
                    self.logger.info(msg)
                    continue

                if given_table_name:
                    # check if the given table_name is in the file name
                    if given_table_name not in file_name_p.name:
                        continue
                table_name = file_name_p.stem
                # remove the suffix if it exists
                check: list = table_name.split(suffix_separator) if suffix_separator else []
                if len(check) == 2:
                    table_name = check[0]
                    suffix = check[1].lower()
                    if suffix in suffix_to_ignore:
                        continue
                elif is_partial:
                    # if we have a partial import, we need to have a suffix !!!
                    # because if not, it's the file with full data to import only when not partial
                    continue
                if not is_partial and table_name in table_is_done:
                    # if we have a partial import, same table can be imported multiple times
                    continue

                cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';")
                if not cursor.fetchone():
                    self.logger.info(f'Table "{table_name}" does not exist and the data will not be imported from {file_name_p.name}.')
                    continue
                if delete_content:
                    cursor.execute(f"DELETE FROM {table_name} WHERE 1")
                    self.connection.commit()
                try:
                    with open(file_name_p, 'r', newline='', encoding='utf-8') as f:
                        reader = csv.reader(f, dialect='unix')
                        csv_columns = next(reader)
                        # Get column names from database
                        cursor.execute(f"PRAGMA table_info({table_name});")
                        db_columns = [column[1] for column in cursor.fetchall()]
                        # Check if columns match
                        if not is_partial and csv_columns != db_columns:
                            msg = f'Columns in CSV file "{file_name}" do not match columns in table "{table_name}".'
                            result.append(msg)
                            self.logger.info(msg)
                            continue
                        for row in reader:
                            if not row:
                                continue
                            data_dict = dict(zip(csv_columns, row))
                            if self._insert_or_update_row(table_name, data_dict):
                                success_count += 1
                            else:
                                fails_count += 1
                        result.append(f'Import: {file_name}')
                        result.append(f'-> Success: {success_count} Fails:{fails_count}')
                        must_reload = True
                        table_is_done.append(table_name)
                        if given_table_name and given_table_name == table_name:
                            break
                except Exception as error:
                    msg = f'Error while importing table "{table_name}" from CSV file "{file_name}"'
                    result.append(msg)
                    self.logger.warning(f'{msg}: {error!r}')
            self.connection.commit()
            cursor.close()
        return result, must_reload

    def generate_test_data(self, number_of_rows=1) -> None:
        """
        Generate and insert the specified number of fake assets into the 'assets' table.
        :param number_of_rows: The number of fake assets to generate and insert.
        """
        # check if the database version is compatible with the current method
        if not self._check_db_version(DbVersionNum.V2, caller_name=inspect.currentframe().f_code.co_name):
            return
        scraped_ids = []
        fake = Faker()
        for index in range(number_of_rows):
            assets_id = fake.uuid4()
            print(f'creating test asset # {index} with id {assets_id}')
            ue_asset = UEAsset()
            # the order of values must match the order of the fields in csv_data.py/csv_sql_fields dict
            data_list = [
                fake.uuid4(),  # asset_id
                fake.sentence(),  # title
                fake.word(),  # category
                round(random.uniform(0, 5), 1),  # review
                random.randint(0, 1000),  # review_count
                fake.name(),  # author
                fake.text(),  # description
                fake.word(),  # status
                round(random.uniform(1, 100), 2),  # discount_price
                round(random.uniform(0, 100), 2),  # discount_percentage
                random.choice([0, 1]),  # discounted
                random.choice([0, 1]),  # is_new
                random.choice([0, 1]),  # free
                random.choice([0, 1]),  # can_purchase
                random.choice([0, 1]),  # owned
                random.choice([0, 1]),  # obsolete
                fake.word(),  # supported_versions
                fake.word(),  # grab_result
                round(random.uniform(1, 100), 2),  # price
                random.randint(0, 1000),  # old_price
                fake.text(),  # comment
                random.randint(1, 5),  # stars
                random.choice([0, 1]),  # must_buy
                fake.word(),  # test_result
                fake.file_path(),  # installed_folder
                fake.sentence(),  # alternative
                fake.word(),  # origin
                random.choice([0, 1]),  # added_manually
                fake.sentence(),  # custom_attributes
                fake.sentence(),  # page_title
                fake.image_url(),  # thumbnail_url
                fake.url(),  # asset_url
                fake.date_time(),  # creation_date
                fake.date_time(),  # update_date
                fake.date_time(),  # date_added_in_db
                assets_id,  # id
                fake.word(),  # namespace
                fake.uuid4(),  # catalog_item_id
                fake.slug(),  # asset_slug
                'USD',  # currency_code
                fake.text(),  # technical_details
                fake.text(),  # long_description
                [fake.word(), fake.word(), fake.word()],  # tags
                fake.uuid4(),  # comment_rating_id
                fake.uuid4(),  # rating_id
                random.choice([0, 1]),  # is_catalog_item
            ]
            ue_asset.init_from_list(data=data_list)
            self.set_assets(assets=ue_asset.data)
            scraped_ids.append(assets_id)
        content = {
            'date': datetime.datetime.now().strftime(default_datetime_format),
            'mode': 'save',
            'files_count': 0,
            'items_count': number_of_rows,
            'scraped_ids': ','.join(scraped_ids)
        }
        self.save_last_run(content)


if __name__ == "__main__":
    # the following code is just for class testing purposes
    clean_data = True
    read_data_only = True  # if True, the code will not create fake assets, but only read them from the database

    db_folder = path_from_relative_to_absolute('../../../scraping/')
    db_name = path_join(db_folder, 'assets.db')
    check_and_create_path(db_name)
    asset_handler = UEAssetDbHandler(database_name=db_name, reset_database=(clean_data and not read_data_only))

    if read_data_only:
        # Read existing assets
        asset_list = asset_handler.get_assets_data()
        print("Assets:", asset_list)
    elif gui_g.s.testing_switch == 1:
        # Create fake assets
        rows_to_create = 300
        if not clean_data:
            rows_count = asset_handler.get_rows_count()
            print(f"Rows count: {rows_count}")
            rows_to_create -= rows_count
        print(f"Creating {rows_to_create} rows")
        asset_handler.generate_test_data(rows_to_create)

    rows_count = asset_handler.get_rows_count()
    print(f"FINAL Rows count: {rows_count}")



# coding=utf-8
"""
Definition for the types used in this module:
- CSVFieldState: Enum for the state of a field in the database
- CSVFieldType: Enum for the (simplified) type of field in the database
- DbVersionNum: The version of the database or/and class.
"""
from enum import Enum

from UEVaultManager.tkgui.modules import globals as gui_g
from UEVaultManager.tkgui.modules.functions_no_deps import convert_to_bool, convert_to_datetime, convert_to_float, convert_to_int


class CSVFieldState(Enum):
    """
    Enum for the state of a field in the database
    Used for filtering the fields regarding the context.
    """
    CSV_ONLY = 0  # field is only in the CSV result file
    SQL_ONLY = 1  # field is only in the database
    CHANGED = 2  # Changed during the process. Will be preserved if already present in data (CSV file or database)
    NOT_PRESERVED = 3  # value will NOT be preserved if already present in data (CSV file or database)
    ASSET_ONLY = 4  # field is only in the property of the UEAsset class
    USER = 5  # field is only in the user data. Will be preserved if already present in data (CSV file or database)


class CSVFieldType(Enum):
    """
    Enum for the (simplified) type of field in the database
    Used for selecting the good format and/or control to show the value.
    """
    STR = 0  # short text (ie ttk.ENTRY)
    INT = 1
    FLOAT = 2
    BOOL = 3
    TEXT = 4  # long text (ie ttk.TEXT)
    DATETIME = 5  # date and time (see ttkbootstrap.widgets.DateEntry)
    LIST = 6  # list of values (ie ttk.DROPBOX)

    def cast(self, value):
        """
        Cast the value to the type of the field.
        :param value: value to cast.
        :return: value with the cast type.
        """
        if self == self.INT:
            return convert_to_int(value)
        if self == self.FLOAT:
            return convert_to_float(value)
        if self == self.BOOL:
            return convert_to_bool(value)
        if self == self.BOOL:
            return convert_to_datetime(value, formats_to_use=[gui_g.s.epic_datetime_format, gui_g.s.csv_datetime_format])
        return str(value)


class DbVersionNum(Enum):
    """
    The version of the database or/and class.
    Used when checking if database must be upgraded by comparing with the class version.
    """
    # when a new version is added to the DbVersionNum enum
    # - add code for the new version to the create_tables() method
    # - add code for the new version check to the check_and_upgrade_database() method
    V0 = 0  # invalid version
    V1 = 1  # initial version : only the "standard" marketplace columns
    V2 = 2  # add the columns used fo user data to the "standard" marketplace columns
    V3 = 3  # add the last_run table to get data about the last run of the app
    V4 = 4  # add custom_attributes field to the assets table
    V5 = 5  # add added_manually column to the assets table
    V6 = 6  # add tags column to the assets table
    V7 = 7  # add the tags table
    V8 = 8  # add the ratings tags table
    V9 = 9  # create the "assets_tags" view for the tags in the assets table
    V10 = 10  # add an autoincrement id to the last_run table
    V11 = 11  # future version



# coding=utf-8
"""
CSV and SQL fields mapping and utility functions.
"""
from datetime import datetime

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
from UEVaultManager.models.types import CSVFieldState, CSVFieldType
from UEVaultManager.tkgui.modules.functions_no_deps import convert_to_bool, convert_to_datetime, convert_to_float, convert_to_int, create_uid

csv_sql_fields = {
    # fields mapping from csv to sql
    # key: csv field name, value: {sql name, state }
    # some field are intentionnaly duplicated because
    #   several CSV fields could come from a same database field
    #   a csv field with this name must exist to get the value
    'Asset_id': {
        'sql_name': 'asset_id',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'App name': {
        'sql_name': 'title',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'App title': {
        # intentionnaly duplicated
        'sql_name': 'title',
        'state': CSVFieldState.CSV_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Category': {
        'sql_name': 'category',
        'state': CSVFieldState.CHANGED,
        'field_type': CSVFieldType.LIST
    },
    'Review': {
        'sql_name': 'review',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.FLOAT
    },
    'Review count': {
        # not in "standard/result" csv file
        'sql_name': 'review_count',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.INT
    },
    'Developer': {
        'sql_name': 'author',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'Description': {
        'sql_name': 'description',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.TEXT
    },
    'Status': {
        'sql_name': 'status',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'Discount price': {
        'sql_name': 'discount_price',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.FLOAT
    },
    'Discount percentage': {
        'sql_name': 'discount_percentage',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.INT
    },
    'Discounted': {
        'sql_name': 'discounted',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.BOOL
    },
    'Is new': {
        # not in "standard/result" csv file
        'sql_name': 'is_new',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.BOOL
    },
    'Free': {
        # not in "standard/result" csv file
        'sql_name': 'free',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.BOOL
    },
    'Can purchase': {
        # not in "standard/result" csv file
        'sql_name': 'can_purchase',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.BOOL
    },
    'Owned': {
        'sql_name': 'owned',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.BOOL
    },
    'Obsolete': {
        'sql_name': 'obsolete',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.BOOL
    },
    'Supported versions': {
        'sql_name': 'supported_versions',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'Grab result': {
        'sql_name': 'grab_result',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.LIST
    },
    'Price': {
        'sql_name': 'price',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.FLOAT
    },
    # ## User Fields
    'Old price': {
        'sql_name': 'old_price',
        'state': CSVFieldState.CHANGED,
        'field_type': CSVFieldType.FLOAT
    },
    'Comment': {
        'sql_name': 'comment',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.TEXT
    },
    'Stars': {
        'sql_name': 'stars',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.INT
    },
    'Must buy': {
        'sql_name': 'must_buy',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.BOOL
    },
    'Test result': {
        'sql_name': 'test_result',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.STR
    },
    'Installed folder': {
        'sql_name': 'installed_folder',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.STR
    },
    'Alternative': {
        'sql_name': 'alternative',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.STR
    },
    'Origin': {
        'sql_name': 'origin',
        'state': CSVFieldState.CHANGED,
        'field_type': CSVFieldType.STR
    },
    'Added manually': {
        'sql_name': 'added_manually',
        'state': CSVFieldState.USER,
        'field_type': CSVFieldType.BOOL
    },
    # ## less important fields
    'Custom attributes': {
        # not in "standard/result" csv file
        'sql_name': 'custom_attributes',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Page title': {
        'sql_name': 'page_title',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'Image': {
        'sql_name': 'thumbnail_url',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    'Url': {
        'sql_name': 'asset_url',
        'state': CSVFieldState.CHANGED,
        'field_type': CSVFieldType.STR
    },
    'Compatible versions': {
        # not in database
        'sql_name': None,
        'state': CSVFieldState.CSV_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Date added': {
        'sql_name': 'date_added_in_db',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.DATETIME
    },
    'Creation date': {
        'sql_name': 'creation_date',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.DATETIME
    },
    'Update date': {
        'sql_name': 'update_date',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.DATETIME
    },
    'UE version': {
        # not in database
        'sql_name': None,
        'state': CSVFieldState.CSV_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Uid': {
        'sql_name': 'id',
        'state': CSVFieldState.NOT_PRESERVED,
        'field_type': CSVFieldType.STR
    },
    # ## UE asset class field only
    'Namespace': {
        'sql_name': 'namespace',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Catalog itemid': {
        'sql_name': 'catalog_item_id',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Asset slug': {
        'sql_name': 'asset_slug',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.STR
    },
    'urlSlug': {
        # intentionnaly duplicated
        'sql_name': 'asset_slug',
        'state': CSVFieldState.CHANGED,
        'field_type': CSVFieldType.STR
    },
    'Currency code': {
        'sql_name': 'currency_code',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Technical details': {
        'sql_name': 'technical_details',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Long description': {
        'sql_name': 'long_description',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.TEXT
    },
    'Tags': {
        'sql_name': 'tags',
        'state': CSVFieldState.SQL_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Comment rating id': {
        'sql_name': 'comment_rating_id',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Rating id': {
        'sql_name': 'rating_id',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
    'Is catalog item': {
        'sql_name': 'is_catalog_item',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.BOOL
    },
    'Thumbnail': {
        # intentionnaly duplicated
        'sql_name': 'thumbnail_url',
        'state': CSVFieldState.ASSET_ONLY,
        'field_type': CSVFieldType.STR
    },
}


def get_csv_field_name_list(exclude_sql_only=True, include_asset_only=False, return_as_string=False, filter_on_states=None):
    """
    Get the csv fields list.
    :param exclude_sql_only: Whether to exclude the sql only fields from result.
    :param include_asset_only: Whether to include the asset only fields from result.
    :param return_as_string: Whether to return a string instead of a list.
    :param filter_on_states: if not empty, only return the fields in the given states.
    :return: csv headings.
    """
    result = []
    for csv_field, value in csv_sql_fields.items():
        if exclude_sql_only and value['state'] == CSVFieldState.SQL_ONLY:
            continue
        if not include_asset_only and value['state'] == CSVFieldState.ASSET_ONLY:
            continue
        if filter_on_states and value['state'] not in filter_on_states:
            continue
        result.append(csv_field)
    if return_as_string:
        result = ','.join(result)
    return result


def get_sql_field_name_list(exclude_csv_only=True, include_asset_only=False, return_as_string=False, add_alias=False, filter_on_states=None):
    """
    Get the sql fields list.
    :param exclude_csv_only: Whether to exclude the csv only fields from result.
    :param include_asset_only: Whether to include the asset only fields from result.
    :param return_as_string: Whether to return a string instead of a list.
    :param add_alias: Whether to add the csv name as alias to the sql field name.
    :param filter_on_states: if not empty, only return the fields in the given states.
    :return: sql headings.
    """
    result = []
    for csv_field, value in csv_sql_fields.items():
        if exclude_csv_only and value['state'] == CSVFieldState.CSV_ONLY:
            continue
        if not include_asset_only and value['state'] == CSVFieldState.ASSET_ONLY:
            continue
        if filter_on_states and value['state'] not in filter_on_states:
            continue
        sql_name = value['sql_name']
        if add_alias:
            if ' AS ' in sql_name:
                result.append(sql_name)
            else:
                result.append(f"{sql_name} AS '{csv_field}'")
        else:
            result.append(sql_name)

    if return_as_string:
        result = ','.join(result)
    return result


def get_typed_value(csv_field='', sql_field='', value='') -> (any, ):
    """
    Get the typed value for a field in CSV or SQL format. Only one of the two fields is required.
    :param csv_field: name of the field (csv format).
    :param sql_field: name of the field (sql format).
    :param value: value to cast.
    :return: typed value.
    """
    if sql_field and not csv_field:
        csv_field = get_csv_field_name(sql_field)

    # noinspection PyBroadException
    try:
        associated_field = csv_sql_fields.get(csv_field, None)
        if associated_field is not None:
            field_type = associated_field['field_type']
            typed_value = field_type.cast(value)
            return typed_value
    except Exception:
        # print(f'Failed to cast value {value}')
        return value
    return value


def is_on_state(csv_field_name: str, states: list[CSVFieldState], default=False) -> bool:
    """
    Check if the csv field is in the given states.
    :param csv_field_name: csv field name.
    :param states: list of states to check.
    :param default: default value if the field is not in the list.
    :return: True if is in the given states.
    """
    if not isinstance(states, list):
        states = [states]
    try:
        state = get_state(csv_field_name)
        return state in states
    except KeyError:
        # print(f'Key not found {csv_field_name} in is_on_state()')  # debug only. Will flood the console
        return default  # by default, we consider that the field is not on this state


def is_from_type(csv_field_name: str, types: list[CSVFieldType], default=False) -> bool:
    """
    Check if the csv field is in the given types.
    :param csv_field_name: csv field name.
    :param types: list of types to check.
    :param default: default value if the field is not in the list.
    :return: True if is in the given types.
    """
    if not isinstance(types, list):
        types = [types]
    try:
        field_type = get_type(csv_field_name)
        return field_type in types
    except KeyError:
        # print(f'Key not found {csv_field_name} in is_from_type()')  # debug only. Will flood the console
        return default  # by default, we consider that the field is not on this type


def get_type(csv_field_name: str):
    """
    Get the type of field.
    :param csv_field_name: csv field name.
    :return: type of the field.
    """
    try:
        return csv_sql_fields[csv_field_name]['field_type']
    except KeyError:
        return None


def get_converters(csv_field_name: str):
    """
    Get the converter of field.
    :param csv_field_name: csv field name.
    :return: list of converters to use sequentially. [str] will be returned if the field is not found.
    """
    field_type = get_type(csv_field_name)

    if field_type == CSVFieldType.LIST:
        # this is a special case. Use a 'category' for pandas datatable.
        # The caller should handle this case where the converter is not callable
        return ['category']
    if field_type == CSVFieldType.INT:
        return [convert_to_int, int]
    if field_type == CSVFieldType.FLOAT:
        return [convert_to_float, float]
    if field_type == CSVFieldType.BOOL:
        return [convert_to_bool, bool]
    if field_type == CSVFieldType.DATETIME:
        return [lambda x: convert_to_datetime(x, formats_to_use=[gui_g.s.epic_datetime_format, gui_g.s.csv_datetime_format])]
    else:
        return [str]


def get_default_value(csv_field_name: str = '', sql_field_name: str = ''):
    """
    Get the default value of field.
    :param csv_field_name: csv field name.
    :param sql_field_name: sql field name. If not provided, the csv field name will be used.
    :return: default value of the field.
    """
    if sql_field_name and not csv_field_name:
        csv_field_name = get_csv_field_name(sql_field_name)
    field_type = get_type(csv_field_name)
    default_values = {
        # CSVFieldType.LIST: [],
        # CSVFieldType.STR: '',
        # CSVFieldType.TEXT: '',
        CSVFieldType.INT: 0,
        CSVFieldType.FLOAT: 0.0,
        CSVFieldType.BOOL: False,
        CSVFieldType.DATETIME: datetime.now().strftime(gui_g.s.csv_datetime_format),
    }
    return default_values.get(field_type, 'None')


def get_state(csv_field_name: str):
    """
    Get the state of field.
    :param csv_field_name: csv field name.
    :return: state of the field.
    """
    try:
        return csv_sql_fields[csv_field_name]['state']
    except KeyError:
        return None


def is_preserved(csv_field_name: str) -> bool:
    """
    Check if the csv field is preserved.
    :param csv_field_name: csv field name.
    :return: True if is preserved.
    """
    return not is_on_state(csv_field_name, [CSVFieldState.NOT_PRESERVED])


def get_sql_field_name(csv_field_name: str):
    """
    Get the sql field name for a csv field name.
    :param csv_field_name: csv field name.
    :return: sql field name.
    """
    try:
        return csv_sql_fields[csv_field_name]['sql_name']
    except KeyError:
        return None


def get_csv_field_name(sql_field_name: str) -> str:
    """
    Get the csv field name for a sql field name.
    :param sql_field_name: sql field name.
    :return: csv field name.
    """
    result = None
    for key, value in csv_sql_fields.items():
        if value['sql_name'] == sql_field_name:
            result = key
            break
    return result


def create_empty_csv_row(return_as_string=False):
    """
    Create an empty row for the csv file.
    :return: empty row.
    """
    data = {}
    uid = create_uid()
    for key in get_csv_field_name_list():
        data[key] = get_default_value(csv_field_name=key)
    data['Asset_id'] = gui_g.s.empty_row_prefix + uid  # dummy unique Asset_id to avoid issue
    data['Image'] = gui_g.s.empty_cell  # avoid displaying image warning on mouse over
    data['Added Manually'] = True
    data['Uid'] = uid
    if return_as_string:
        data = ','.join(str(value) for value in data.values())
    return data


def convert_csv_row_to_sql_row(csv_row: dict) -> dict:
    """
    Convert a csv row to a sql row.
    :param csv_row: csv row.
    :return: sql row.
    """
    sql_row = {}
    for csv_field, value in csv_row.items():
        sql_field = get_sql_field_name(csv_field)
        if sql_field:
            sql_row[sql_field] = value
    return sql_row



# coding: utf-8
"""
implementation for:
- EPCLFS: Epic Games Encrypted Config Filesystem.
"""
import configparser
import os

from UEVaultManager.lfs.utils import path_join


class EPCLFS:
    """
    Epic Games Encrypted Config Filesystem Class.
    """
    # Known encryption key(s) for JSON user data
    # encrypted using AES-256-ECB mode
    data_keys = []

    def __init__(self):
        if os.name == 'nt':
            self.appdata_path = os.path.expandvars(r'%LOCALAPPDATA%\EpicGamesLauncher\Saved\Config\Windows')
            self.programdata_path = os.path.expandvars(r'%PROGRAMDATA%\Epic\EpicGamesLauncher\Data\Manifests')
        else:
            self.appdata_path = self.programdata_path = None

        self.config = configparser.ConfigParser(strict=False)
        self.config.optionxform = lambda option: option

        self.manifests = dict()

    def read_config(self):
        """
        Read the EGS config files.
        """
        if not self.appdata_path:
            raise ValueError('EGS AppData path is not set')

        if not os.path.isdir(self.appdata_path):
            raise ValueError('EGS AppData path does not exist')

        self.config.read(path_join(self.appdata_path, 'GameUserSettings.ini'), encoding='utf-8')
